"""pipeline.py unit tests.

A lot of the tests for the Pipeline.new_pipe_and_args factory constructor
exist in ./pipelinerunner_test.py, which tests at a higher level that the
inputs from a run request map as expected into the Pipeline instance.
"""
import logging
from unittest.mock import call, patch, Mock

import pytest

from pypyr.cache.loadercache import loader_cache
from pypyr.cache.parsercache import contextparser_cache
from pypyr.context import Context
from pypyr.errors import (ContextError,
                          KeyNotInContextError,
                          PyModuleNotFoundError,
                          Stop,
                          StopPipeline,
                          StopStepGroup)
from pypyr.pipeline import Pipeline
from pypyr.pipedef import PipelineDefinition, PipelineInfo
from tests.common.utils import DeepCopyMagicMock

from tests.common.utils import patch_logger


def get_pipe_def(dict_in, info=None):
    """Wrap input dict & info into a PipelineDefinition."""
    return PipelineDefinition(pipeline=dict_in, info=info)

# region context parser
# region parser mocks


def mock_parser_arb(args):
    """Arbitrary mock function to execute instead of get_parsed_context."""
    return Context({'key1': 'created in mock parser', 'key2': args})


def mock_parser_none(args):
    """Return None, mocking get_parsed_context."""
    return None
# endregion parser mocks


@patch('pypyr.cache.loadercache.Loader.get_pipeline')
def test_get_parsed_context_no_parser(mock_get_pipeline):
    """On get_parsed_context return empty Context when no parser specified."""
    mock_get_pipeline.return_value = get_pipe_def({})
    context = Context()
    pipeline = Pipeline('arb')
    pipeline.run(context)

    assert context == {}
    mock_get_pipeline.assert_called_once_with(name='arb', parent=None)


@patch('pypyr.cache.loadercache.Loader.get_pipeline')
def test_get_parsed_context_parser_not_found(mock_get_pipeline):
    """On get_parsed_context raise if parser module specified but not found."""
    mock_get_pipeline.return_value = get_pipe_def({
        'context_parser': 'unlikelyblahmodulenameherexxssz'})
    context = Context()
    pipeline = Pipeline('arb')

    with pytest.raises(PyModuleNotFoundError):
        pipeline.run(context)

    assert context == {}
    mock_get_pipeline.assert_called_once_with(name='arb', parent=None)


@patch('pypyr.moduleloader.get_module')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
def test_get_parsed_context_parser_returns_none(mock_get_pipeline,
                                                mock_moduleloader):
    """On get_parsed_context return empty Context when parser returns None."""
    mock_moduleloader.return_value.get_parsed_context = mock_parser_none
    mock_get_pipeline.return_value = get_pipe_def(
        {'context_parser': 'specifiedparserhere'})

    pipeline = Pipeline('arb', context_args=['in arg here'])

    context = Context()

    pipeline.run(context)

    mock_moduleloader.assert_called_once_with('specifiedparserhere')
    mock_get_pipeline.assert_called_once_with(name='arb', parent=None)
    assert context == {}


@patch('pypyr.moduleloader.get_module')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
def test_get_parsed_context_parser_pass(mock_get_pipeline, mock_moduleloader):
    """On get_parsed_context pass arg param and returns context."""
    contextparser_cache.clear()
    mock_moduleloader.return_value.get_parsed_context = mock_parser_arb
    mock_get_pipeline.return_value = get_pipe_def(
        {'context_parser': 'specifiedparserhere'})

    pipeline = Pipeline('arb', context_args='in arg here')

    context = Context()

    pipeline.run(context)

    mock_moduleloader.assert_called_once_with('specifiedparserhere')
    mock_get_pipeline.assert_called_once_with(name='arb', parent=None)

    assert isinstance(context, Context)
    assert len(context) == 2
    assert context['key1'] == 'created in mock parser'
    assert context['key2'] == 'in arg here'


@patch('pypyr.moduleloader.get_module', return_value=3)
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
def test_get_parser_context_signature_wrong(mock_get_pipeline,
                                            mock_moduleloader):
    """Raise when parser found but no get_parsed_context attr."""
    contextparser_cache.clear()

    mock_get_pipeline.return_value = get_pipe_def(
        {'context_parser': 'specifiedparserhere'})

    pipeline = Pipeline('arb', context_args='in arg here')

    context = Context()

    with pytest.raises(AttributeError) as err_info:
        pipeline.run(context)

    mock_moduleloader.assert_called_once_with('specifiedparserhere')
    mock_get_pipeline.assert_called_once_with(name='arb', parent=None)

    assert str(err_info.value) == ("'int' object has no attribute "
                                   "'get_parsed_context'")


@patch('pypyr.moduleloader.get_module')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
def test_prepare_context_empty_parse(mock_get_pipeline,
                                     mock_moduleloader):
    """Empty parsed_context works."""
    contextparser_cache.clear()
    parser = Mock()
    parser.return_value = {}
    mock_moduleloader.return_value.get_parsed_context = parser
    mock_get_pipeline.return_value = get_pipe_def(
        {'context_parser': 'specifiedparserhere'})

    pipeline = Pipeline('arb', context_args='arb context input')

    context = Context({'c1': 'cv1', 'c2': 'cv2'})

    pipeline.run(context)

    mock_moduleloader.assert_called_once_with('specifiedparserhere')
    parser.assert_called_once_with('arb context input')
    mock_get_pipeline.assert_called_once_with(name='arb', parent=None)

    assert context == {'c1': 'cv1', 'c2': 'cv2'}


@patch('pypyr.moduleloader.get_module')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
def test_prepare_context_with_parse_merge(mock_get_pipeline,
                                          mock_moduleloader):
    """On parsed_context override context."""
    contextparser_cache.clear()
    parser = Mock()
    parser.return_value = {'a': 'av1', 'c1': 'new value from parsed'}
    mock_moduleloader.return_value.get_parsed_context = parser
    mock_get_pipeline.return_value = get_pipe_def(
        {'context_parser': 'specifiedparserhere'})

    pipeline = Pipeline('arb', context_args='arb context input')

    context = Context({'c1': 'cv1', 'c2': 'cv2'})
    pipeline.run(context)

    mock_moduleloader.assert_called_once_with('specifiedparserhere')
    parser.assert_called_once_with('arb context input')
    mock_get_pipeline.assert_called_once_with(name='arb', parent=None)

    assert context == {'a': 'av1', 'c1': 'new value from parsed', 'c2': 'cv2'}

# endregion context parser

# region loader


def test_arbitrary_loader_module_not_found():
    """Raise when loader not found."""
    loader_cache.clear()

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        loader='not_found_loader')

    with pytest.raises(PyModuleNotFoundError):
        pipeline.run(Context())


def test_loader_no_get_pipeline_definition():
    """Arbitrary loader module without `get_pipeline_definition` function."""
    loader_cache.clear()
    import sys
    current_module = sys.modules[__name__]

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        loader=__name__)
    with patch_logger(
            'pypyr.cache.loadercache',
            logging.ERROR) as mock_logger_error:
        with pytest.raises(AttributeError) as err:
            pipeline.run(Context())

    assert str(err.value) == f"module '{__name__}' " \
        "has no attribute 'get_pipeline_definition'"

    mock_logger_error.assert_called_once_with(
        f"The pipeline loader {current_module} doesn't have a "
        "get_pipeline_definition(pipeline_name, parent) function."
    )


@patch('pypyr.loaders.file.get_pipeline_definition')
def test_empty_loader_set_up_to_default(mock_get_pipeline_definition):
    """Default loader should be pypyr.loaders.file."""
    loader_cache.clear()

    mock_get_pipeline_definition.return_value = get_pipe_def({'steps': None})

    pipeline = Pipeline('arb pipe', context_args='arb context input')
    pipeline.run(Context())

    mock_get_pipeline_definition.assert_called_once_with(
        pipeline_name='arb pipe',
        parent=None
    )


@patch('pypyr.loaders.file.get_pipeline_definition')
def test_empty_loader_set_up_to_default_with_parent(
        mock_get_pipeline_definition):
    """Default loader should be pypyr.loaders.file with parent."""
    loader_cache.clear()

    mock_get_pipeline_definition.return_value = get_pipe_def({'steps': None})

    pipeline = Pipeline('arb pipe', context_args='arb context input')
    pipeline.load_and_run_pipeline(Context(), parent='/arb/dir')

    mock_get_pipeline_definition.assert_called_once_with(
        pipeline_name='arb pipe',
        parent='/arb/dir'
    )


def test_arb_loader():
    """Test loader set up."""
    loader_cache.clear()

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        loader='arbpack.arbloader',
                        py_dir='tests')
    pipeline.load_and_run_pipeline(Context(), parent='/arb/dir')

    loader = loader_cache.get_pype_loader('arbpack.arbloader')
    assert loader.name == 'arbpack.arbloader'
    assert loader.get_pipeline('arb pipe', '/arb/dir').pipeline == {
        'pipeline_name': 'arb pipe', 'parent': '/arb/dir'}

    loader_cache.clear()


def test_arb_loader_no_parent():
    """Test loader set up with no parent."""
    loader_cache.clear()

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        loader='arbpack.arbloader',
                        py_dir='tests')
    pipeline.load_and_run_pipeline(Context())

    loader = loader_cache.get_pype_loader('arbpack.arbloader')
    assert loader.name == 'arbpack.arbloader'
    assert loader.get_pipeline('arb pipe', None).pipeline == {
        'pipeline_name': 'arb pipe', 'parent': None}

    loader_cache.clear()

# endregion loader

# region run_pipeline


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_pass_minimal(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Create implicit context if doesn't exist & no context parser."""
    pipe_def = get_pipe_def({'arb': 'pipe'})
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.return_value = {'a': 'b'}
    mock_parser.return_value = parser

    pipeline = Pipeline('arb pipe', context_args='arb context input')
    context_instance = Context()
    with patch('pypyr.pipeline.Context') as mock_context:
        mock_context.return_value = context_instance
        pipeline.load_and_run_pipeline(None)

    mock_add_sys_path.assert_not_called()
    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent=None)

    mock_parser.assert_not_called()
    parser.assert_not_called()

    mock_context.assert_called_once()

    # assure that stack empty when done
    assert not context_instance._stack

    mock_steps_runner.assert_called_once_with(
        pipeline_body={'arb': 'pipe'},
        context=context_instance)
    # No called steps
    sr = mock_steps_runner.return_value
    sr.run_step_groups.assert_called_once_with(groups=['steps'],
                                               success_group='on_success',
                                               failure_group='on_failure')
    sr.run_failure_step_group.assert_not_called()


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_pass_skip_parse_context(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Explicit False parse_input doesn't run parser."""
    pipe_def = get_pipe_def({'arb': 'pipe'})
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.return_value = {'a': 'b'}
    mock_parser.return_value = parser

    context = Context({'c': 'd'})
    pipeline = Pipeline('arb pipe', parse_input=False)
    pipeline.load_and_run_pipeline(context)

    mock_add_sys_path.assert_not_called()
    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent=None)
    mock_parser.assert_not_called()
    parser.assert_not_called()

    mock_steps_runner.assert_called_once_with(
        pipeline_body={'arb': 'pipe'},
        context=context)
    # No called steps, just on_failure since err on parse context already
    sr = mock_steps_runner.return_value
    sr.run_step_groups.assert_called_once_with(groups=['steps'],
                                               success_group='on_success',
                                               failure_group='on_failure')
    sr.run_failure_step_group.assert_not_called()


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_parse_context_error(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """run_pipeline on_failure with Context as is if parse fails."""
    pipe_def = get_pipe_def({'context_parser': 'arb parser'})
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.side_effect = ContextError
    mock_parser.return_value = parser

    context = Context({'c': 'd'})
    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        parse_input=True)

    with pytest.raises(ContextError):
        pipeline.load_and_run_pipeline(context)

    assert context == {'c': 'd'}
    mock_add_sys_path.assert_not_called()
    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent=None)

    mock_parser.assert_called_once_with('arb parser')
    parser.assert_called_once_with('arb context input')

    mock_steps_runner.assert_called_once_with(
        pipeline_body=pipe_def.pipeline,
        context=context)
    # No called steps, just on_failure since err on parse context already
    sr = mock_steps_runner.return_value
    sr.run_step_groups.assert_not_called()
    sr.run_failure_step_group.assert_called_once_with('on_failure')


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_steps_error_raises(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Run on_failure and raise error if steps group fails."""
    # First time it runs is steps - give a KeyNotInContextError.
    pipe_def = get_pipe_def({'context_parser': 'arb parser'})
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.return_value = {'a': 'b'}
    mock_parser.return_value = parser

    context = Context({'c': 'd'})
    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        parse_input=True)

    mock_steps_runner.return_value.run_step_groups.side_effect = (
        KeyNotInContextError)

    with pytest.raises(KeyNotInContextError):
        pipeline.run(context)

    assert context == {'a': 'b', 'c': 'd'}

    mock_add_sys_path.assert_not_called()
    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent=None)

    mock_parser.assert_called_once_with('arb parser')
    parser.assert_called_once_with('arb context input')

    mock_steps_runner.return_value.run_step_groups.assert_called_once_with(
        groups=['steps'],
        success_group='on_success',
        failure_group='on_failure'
    )
    mock_steps_runner.assert_called_once_with(
        pipeline_body=pipe_def.pipeline,
        context=context)


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_with_existing_context_pass(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Pipeline runs with existing context."""
    pipe_def = get_pipe_def({'context_parser': 'arb parser'})
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.return_value = {'1': 'context 1', '2': 'context2'}
    mock_parser.return_value = parser

    context = Context({'2': 'original', '3': 'new'})

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input')
    pipeline.load_and_run_pipeline(context)

    assert not context.is_in_pipeline_scope
    mock_add_sys_path.assert_not_called()
    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent=None)

    mock_parser.assert_called_once_with('arb parser')
    parser.assert_called_once_with('arb context input')

    mock_steps_runner.return_value.run_step_groups.assert_called_once_with(
        groups=['steps'],
        success_group='on_success',
        failure_group='on_failure'
    )
    mock_steps_runner.assert_called_once_with(
        pipeline_body={'context_parser': 'arb parser'},
        context={'1': 'context 1',
                 '2': 'context2',
                 '3': 'new'})


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_with_dir_specified(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Py dir passed to add_sys_path."""
    pipe_yaml = {'context_parser': 'arb parser'}
    pipe_def = get_pipe_def(pipe_yaml)
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.return_value = {'1': 'context 1', '2': 'context2'}
    mock_parser.return_value = parser

    context = Context({'2': 'original', '3': 'new'})

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        py_dir='/arb/dir')
    pipeline.load_and_run_pipeline(context)

    assert not context.is_in_pipeline_scope
    mock_add_sys_path.assert_called_once_with('/arb/dir')
    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent=None)

    mock_parser.assert_called_once_with('arb parser')
    parser.assert_called_once_with('arb context input')

    mock_steps_runner.return_value.run_step_groups.assert_called_once_with(
        groups=['steps'],
        success_group='on_success',
        failure_group='on_failure'
    )
    mock_steps_runner.assert_called_once_with(pipeline_body=pipe_yaml,
                                              context={'1': 'context 1',
                                                       '2': 'context2',
                                                       '3': 'new'})


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_with_group_specified(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Run pipeline with specified groups."""
    pipe_yaml = {'arb': 'pipe'}
    pipe_def = get_pipe_def(pipe_yaml)
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.return_value = {'1': 'context 1', '2': 'context2'}
    mock_parser.return_value = parser

    context = Context({'2': 'original', '3': 'new'})

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        groups=['arb1', 'arb2'])

    pipeline.load_and_run_pipeline(context)

    assert not context.is_in_pipeline_scope
    mock_add_sys_path.assert_not_called()
    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent=None)

    mock_parser.assert_not_called()
    parser.assert_not_called()

    mock_steps_runner.return_value.run_step_groups.assert_called_once_with(
        groups=['arb1', 'arb2'],
        success_group=None,
        failure_group=None
    )
    mock_steps_runner.assert_called_once_with(pipeline_body=pipe_yaml,
                                              context={'2': 'original',
                                                       '3': 'new'})


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_with_parent_specified(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Run pipeline with specified parent."""
    pipe_yaml = {'arb': 'pipe'}
    pipe_def = get_pipe_def(pipe_yaml)
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.return_value = {'1': 'context 1', '2': 'context2'}
    mock_parser.return_value = parser

    context = Context({'2': 'original', '3': 'new'})

    pipeline = Pipeline('arb pipe')

    pipeline.load_and_run_pipeline(context, '/parent')

    assert not context.is_in_pipeline_scope
    mock_add_sys_path.assert_not_called()
    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent='/parent')

    mock_parser.assert_not_called()
    parser.assert_not_called()

    mock_steps_runner.return_value.run_step_groups.assert_called_once_with(
        groups=['steps'],
        success_group='on_success',
        failure_group='on_failure'
    )
    mock_steps_runner.assert_called_once_with(pipeline_body=pipe_yaml,
                                              context={'2': 'original',
                                                       '3': 'new'})


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_with_success_group_specified(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Run pipeline with specified success group."""
    pipe_yaml = {'context_parser': 'arb parser'}
    pipe_def = get_pipe_def(pipe_yaml)
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.return_value = {'1': 'context 1', '2': 'context2'}
    mock_parser.return_value = parser

    context = Context({'2': 'original', '3': 'new'})

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        success_group='arb1')

    pipeline.load_and_run_pipeline(context)

    assert not context.is_in_pipeline_scope
    mock_add_sys_path.assert_not_called()

    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent=None)

    mock_parser.assert_called_once_with('arb parser')
    parser.assert_called_once_with('arb context input')

    mock_steps_runner.return_value.run_step_groups.assert_called_once_with(
        groups=['steps'],
        success_group='arb1',
        failure_group=None
    )
    mock_steps_runner.assert_called_once_with(pipeline_body=pipe_yaml,
                                              context={'1': 'context 1',
                                                       '2': 'context2',
                                                       '3': 'new'})


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_with_failure_group_specified(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Run pipeline with specified failure group."""
    pipe_yaml = {'context_parser': 'arb parser'}
    pipe_def = get_pipe_def(pipe_yaml)
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.return_value = {'1': 'context 1', '2': 'context2'}
    mock_parser.return_value = parser

    context = Context({'2': 'original', '3': 'new'})

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        failure_group='arb1')

    pipeline.load_and_run_pipeline(context)

    assert not context.is_in_pipeline_scope
    mock_add_sys_path.assert_not_called()

    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent=None)

    mock_parser.assert_called_once_with('arb parser')
    parser.assert_called_once_with('arb context input')

    mock_steps_runner.return_value.run_step_groups.assert_called_once_with(
        groups=['steps'],
        success_group=None,
        failure_group='arb1'
    )
    mock_steps_runner.assert_called_once_with(pipeline_body=pipe_yaml,
                                              context={'1': 'context 1',
                                                       '2': 'context2',
                                                       '3': 'new'})


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_load_and_run_pipeline_with_group_and_failure_group_specified(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Pass run_pipeline with specified group and failure group."""
    pipe_yaml = {'context_parser': 'arb parser'}
    pipe_def = get_pipe_def(pipe_yaml)
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.return_value = {'1': 'context 1', '2': 'context2'}
    mock_parser.return_value = parser

    context = Context({'2': 'original', '3': 'new'})

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        groups=['arb1'],
                        failure_group='arb2')

    pipeline.load_and_run_pipeline(context)

    assert not context.is_in_pipeline_scope
    mock_add_sys_path.assert_not_called()

    mock_parser.assert_called_once_with('arb parser')
    parser.assert_called_once_with('arb context input')

    mock_get_pipe.assert_called_once_with(name='arb pipe',
                                          parent=None)

    mock_steps_runner.return_value.run_step_groups.assert_called_once_with(
        groups=['arb1'],
        success_group=None,
        failure_group='arb2'
    )
    mock_steps_runner.assert_called_once_with(pipeline_body=pipe_yaml,
                                              context={'1': 'context 1',
                                                       '2': 'context2',
                                                       '3': 'new'})


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_run_pipeline_parse_context_error_failure(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Run on_failure on context parse exception."""
    pipe_yaml = {'context_parser': 'arb parser'}
    pipe_def = get_pipe_def(pipe_yaml)
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.side_effect = ValueError('arb')
    mock_parser.return_value = parser

    context = Context({'2': 'original', '3': 'new'})

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input',
                        groups=['gr'],
                        success_group='sg',
                        failure_group='fg')

    with pytest.raises(ValueError) as err:
        pipeline.run(context)

    assert str(err.value) == 'arb'

    assert not context.is_in_pipeline_scope
    mock_add_sys_path.assert_not_called()

    mock_parser.assert_called_once_with('arb parser')
    parser.assert_called_once_with('arb context input')

    mock_steps_runner.assert_called_once_with(pipeline_body=pipe_yaml,
                                              context={'2': 'original',
                                                       '3': 'new'})
    # No called steps, just on_failure since err on parse context already
    sr = mock_steps_runner.return_value
    sr.run_step_groups.assert_not_called()
    sr.run_failure_step_group.assert_called_once_with('fg')


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_run_pipeline_parse_context_error_failure_stop(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Run on_failure on context parser exception with Stop."""
    pipe_yaml = {'context_parser': 'arb parser'}
    pipe_def = get_pipe_def(pipe_yaml)
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.side_effect = ValueError('arb')
    mock_parser.return_value = parser

    sr = mock_steps_runner.return_value

    sr.run_failure_step_group.side_effect = Stop()

    context = Context()
    pipeline = Pipeline('arb pipe',
                        context_args='arb context input')

    with pytest.raises(Stop):
        pipeline.load_and_run_pipeline(context)

    assert not context.is_in_pipeline_scope
    mock_add_sys_path.assert_not_called()

    mock_parser.assert_called_once_with('arb parser')
    parser.assert_called_once_with('arb context input')

    mock_steps_runner.assert_called_once_with(pipeline_body=pipe_yaml,
                                              context={})

    # No called steps, just on_failure since err on parse context already
    sr.run_step_groups.assert_not_called()
    sr.run_failure_step_group.assert_called_once_with('on_failure')


@patch('pypyr.pipeline.StepsRunner', autospec=True)
@patch('pypyr.cache.parsercache.contextparser_cache.get_context_parser')
@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.moduleloader.add_sys_path')
def test_run_pipeline_parse_context_error_failure_stopstepgroup(
        mock_add_sys_path,
        mock_get_pipe,
        mock_parser,
        mock_steps_runner):
    """Context failure handler swallows StopStepGroup."""
    pipe_yaml = {'context_parser': 'arb parser'}
    pipe_def = get_pipe_def(pipe_yaml)
    mock_get_pipe.return_value = pipe_def

    parser = Mock()
    parser.side_effect = ValueError('arb')
    mock_parser.return_value = parser

    sr = mock_steps_runner.return_value

    sr.run_failure_step_group.side_effect = StopStepGroup()
    context = Context()

    pipeline = Pipeline('arb pipe',
                        context_args='arb context input')

    with pytest.raises(ValueError) as err:
        pipeline.load_and_run_pipeline(context)

    assert str(err.value) == 'arb'

    assert not context.is_in_pipeline_scope
    mock_add_sys_path.assert_not_called()

    mock_parser.assert_called_once_with('arb parser')
    parser.assert_called_once_with('arb context input')

    mock_steps_runner.assert_called_once_with(pipeline_body=pipe_yaml,
                                              context={})

    # No called steps, just on_failure since err on parse context already
    sr.run_step_groups.assert_not_called()
    sr.run_failure_step_group.assert_called_once_with('on_failure')

# endregion run_pipeline

# region Stop & StopPipeline

# region stop helpers


def get_test_pipeline_definition(pipeline):
    """Wrap input pipeline (dict) into a PipelineDefinition.

    Args:
        pipeline (dict-like): pipeline payload.

    Returns:
        PipelineDefinition with pipeline payload and arb PipelineInfo.
    """
    return PipelineDefinition(
        pipeline=pipeline,
        info=PipelineInfo(pipeline_name='arbpipe',
                          loader='arbloader',
                          parent='arbdir'))


def get_step_pipeline():
    """Test pipeline for jump wrapped in PipelineDefinition."""
    return get_test_pipeline_definition(get_step_pipeline_payload())


def get_step_pipeline_payload():
    """Bare dict pipeline payload."""
    return {
        'sg1': [
            'sg1.step1',
            'sg1.step2'
        ],
        'sg2': [
            'sg2.step1',
            'sg2.step2'
        ],
        'sg3': [
            'sg3.step1',
            'sg3.step2'
        ],
        'sg4': [
            'sg4.step1',
            'sg4.step2'
        ],
        'sg5': [
            'sg5.step1'
        ],
        'sg6': [
            'sg6.step1',
            'sg6.step2'
        ]
    }


def nothing_step(context):
    """Mock step."""
    pass


def stop_pipe_step(context):
    """Mock stop pipeline step."""
    raise StopPipeline()


def stop_all_step(context):
    """Mock stop all step."""
    raise Stop()

# endregion stop helpers


@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_stop_pipeline(mock_step_cache, mock_get_pipe):
    """When StopPipeline stop pipeline execution."""
    # Sequence: sg2 - sg2.1, 2.2
    #           sg3 - sg3.1 (StopPipeline)
    mock_step_cache.side_effect = [
        nothing_step,  # 2.1
        nothing_step,  # 2.2
        stop_pipe_step,  # 3.1
    ]

    mock_get_pipe.return_value = get_test_pipeline_definition(
        get_step_pipeline_payload())
    context = Context()

    pipeline = Pipeline('arb pipe',
                        groups=['sg2', 'sg3', 'sg4', 'sg1'],
                        success_group='sg5',
                        failure_group=None)

    pipeline.run(context)

    assert not context.is_in_pipeline_scope
    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg2.step2'),
                                          call('sg3.step1')
                                          ]


@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_stop_pipeline_for(mock_step_cache, mock_get_pipe):
    """When StopPipeline stop pipeline execution in for loop."""
    # Sequence: sg2 - sg2.1, 2.2
    #           sg3 - sg3.1 x2 (StopPipeline)

    nothing_mock = DeepCopyMagicMock()
    mock312 = DeepCopyMagicMock()

    def step31(context):
        mock312(context)
        if context['i'] == 'two':
            raise StopPipeline()

    mock_step_cache.side_effect = [
        nothing_mock,  # 2.1
        nothing_mock,  # 2.2
        step31,  # 3.1
    ]

    mock_get_pipe.return_value = get_for_step_pipeline()
    context = Context()

    pipeline = Pipeline('arb pipe',
                        groups=['sg2', 'sg3', 'sg4', 'sg1'],
                        success_group='sg5',
                        failure_group=None)

    pipeline.run(context)

    assert not context.is_in_pipeline_scope
    assert nothing_mock.mock_calls == [call({}),
                                       call({})
                                       ]

    assert mock312.mock_calls == [call({'i': 'one'}),
                                  call({'i': 'two'})]

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg2.step2'),
                                          call('sg3.step1')
                                          ]


def get_retry_step_pipeline():
    """Test pipeline for retry loop."""
    return {
        'sg1': [
            'sg1.step1',
            'sg1.step2'
        ],
        'sg2': [
            'sg2.step1',
            'sg2.step2'
        ],
        'sg3': [
            {'name': 'sg3.step1',
             'retry': {'max': 3}
             },
            'sg3.step2'
        ],
        'sg4': [
            'sg4.step1',
            'sg4.step2'
        ],
        'sg5': [
            'sg5.step1'
        ],
        'sg6': [
            'sg6.step1',
            'sg6.step2'
        ]
    }


@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_stop_pipeline_retry(mock_step_cache, mock_get_pipe):
    """When StopPipeline stop pipeline execution in retry loop."""
    # Sequence: sg2 - sg2.1, 2.2
    #           sg3 - sg3.1 x2 (StopPipeline)

    nothing_mock = DeepCopyMagicMock()
    mock312 = DeepCopyMagicMock()

    def step31(context):
        mock312(context)
        if context['retryCounter'] == 2:
            raise StopPipeline()
        else:
            raise ValueError(context['retryCounter'])

    mock_step_cache.side_effect = [
        nothing_mock,  # 2.1
        nothing_mock,  # 2.2
        step31,  # 3.1
    ]

    pipe_yaml = get_retry_step_pipeline()
    mock_get_pipe.return_value = get_test_pipeline_definition(pipe_yaml)
    context = Context()

    pipeline = Pipeline('arb pipe',
                        groups=['sg2', 'sg3', 'sg4', 'sg1'],
                        success_group='sg5',
                        failure_group=None)

    pipeline.run(context)

    assert not context.is_in_pipeline_scope

    assert nothing_mock.mock_calls == [call({}),
                                       call({})
                                       ]

    assert mock312.mock_calls == [call({'retryCounter': 1}),
                                  call({'retryCounter': 2})]

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg2.step2'),
                                          call('sg3.step1')
                                          ]


@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_stop_all(mock_step_cache, mock_get_pipe):
    """Stop stops pipeline execution."""
    # Sequence: sg2 - sg2.1, 2.2
    #           sg3 - sg3.1 (StopPipeline)
    mock_step_cache.side_effect = [
        nothing_step,  # 2.1
        nothing_step,  # 2.2
        stop_all_step,  # 3.1
    ]

    mock_get_pipe.return_value = get_step_pipeline()

    context = Context()

    pipeline = Pipeline('arb pipe',
                        groups=['sg2', 'sg3', 'sg4', 'sg1'],
                        success_group='sg5',
                        failure_group=None)

    pipeline.run(context)

    assert not context.is_in_pipeline_scope

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg2.step2'),
                                          call('sg3.step1')
                                          ]


def get_while_step_pipeline():
    """Test pipeline for while."""
    return get_test_pipeline_definition({
        'sg1': [
            'sg1.step1',
            'sg1.step2'
        ],
        'sg2': [
            'sg2.step1',
            'sg2.step2'
        ],
        'sg3': [
            {'name': 'sg3.step1',
             'while': {
                 'max': 3},
             },
            'sg3.step2'
        ],
        'sg4': [
            'sg4.step1',
            'sg4.step2'
        ],
        'sg5': [
            'sg5.step1'
        ],
        'sg6': [
            'sg6.step1',
            'sg6.step2'
        ]
    })


@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_stop_all_while(mock_step_cache, mock_get_pipe):
    """Stop stops pipeline execution inside a while."""
    # Sequence: sg2 - sg2.1, 2.2
    #           sg3 - sg3.1 loop 3, StopPipeline on 2
    nothing_mock = DeepCopyMagicMock()
    mock312 = DeepCopyMagicMock()

    def step31(context):
        mock312(context)
        if context['whileCounter'] == 2:
            raise Stop()

    mock_step_cache.side_effect = [
        nothing_mock,  # 2.1
        nothing_mock,  # 2.2
        step31  # 3.1.2
    ]
    mock_get_pipe.return_value = get_while_step_pipeline()

    context = Context()

    pipeline = Pipeline('arb pipe',
                        groups=['sg2', 'sg3', 'sg4', 'sg1'],
                        success_group='sg5',
                        failure_group=None)

    pipeline.run(context)

    assert not context.is_in_pipeline_scope

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg2.step2'),
                                          call('sg3.step1')
                                          ]

    assert nothing_mock.mock_calls == [call({}),
                                       call({})
                                       ]

    assert mock312.mock_calls == [call({'whileCounter': 1}),
                                  call({'whileCounter': 2})]


def get_for_step_pipeline():
    """Test pipeline for for loop."""
    return get_test_pipeline_definition(get_for_step_pipeline_payload())


def get_for_step_pipeline_payload():
    """Bare dict for for pipeline."""
    return {
        'sg1': [
            'sg1.step1',
            'sg1.step2'
        ],
        'sg2': [
            'sg2.step1',
            'sg2.step2'
        ],
        'sg3': [
            {'name': 'sg3.step1',
             'foreach': ['one', 'two', 'three']
             },
            'sg3.step2'
        ],
        'sg4': [
            'sg4.step1',
            'sg4.step2'
        ],
        'sg5': [
            'sg5.step1'
        ],
        'sg6': [
            'sg6.step1',
            'sg6.step2'
        ]
    }


@patch('pypyr.cache.loadercache.Loader.get_pipeline')
@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_stop_all_for(mock_step_cache, mock_get_pipe):
    """Stop stops pipeline execution inside a for loop."""
    # Sequence: sg2 - sg2.1, 2.2
    #           sg3 - sg3.1 loop 3, StopPipeline on 2
    nothing_mock = DeepCopyMagicMock()
    mock312 = DeepCopyMagicMock()

    def step31(context):
        mock312(context)
        if context['i'] == 'two':
            raise Stop()

    mock_step_cache.side_effect = [
        nothing_mock,  # 2.1
        nothing_mock,  # 2.2
        step31  # 3.1.2
    ]

    mock_get_pipe.return_value = get_for_step_pipeline()

    context = Context()

    pipeline = Pipeline('arb pipe',
                        groups=['sg2', 'sg3', 'sg4', 'sg1'],
                        success_group='sg5',
                        failure_group=None)

    pipeline.run(context)

    assert not context.is_in_pipeline_scope

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg2.step2'),
                                          call('sg3.step1')
                                          ]

    assert nothing_mock.mock_calls == [call({}),
                                       call({})
                                       ]

    assert mock312.mock_calls == [call({'i': 'one'}),
                                  call({'i': 'two'})]
# endregion Stop & StopPipeline



"""Spawn subprocess synchronously."""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
from collections.abc import Sequence
from contextlib import contextmanager
import logging
from os import PathLike
from pathlib import Path
import shlex
import subprocess

from pypyr.config import config
from pypyr.errors import ContextError, SubprocessError

logger = logging.getLogger(__name__)

SimpleCommandTypes = (str, bytes, PathLike)


class Command:
    """A subprocess run instruction. Use run() to spawn the subprocess.

    Attributes:
        cmd (str/bytes/Pathlike or list[str/bytes/Pathlike]): A str for a
            single run instruction, or a list of str for >1 run instructions.
        is_shell (bool): Invoke subprocess through the default shell.
        cwd (Path-like): Current working directory to this during execution.
        is_save (bool): Save output to self.results.
        is_text (bool): Treat stdin, stdout & stderr streams as text in the
            provided encoding, not bytes. This is only relevant when
            is_save is True.
        stdout (None | str | Pathlike): Write stdout here. None means inherit
            the parent's handle, '/dev/null' means to write to the null device,
            and a path will write to file.
        stderr (None | str | Pathlike): Write stderr here. None means inherit
            the parent's handle, '/dev/null' means to write to the null device,
            '/dev/stdout' redirects to stdout, and a path will write to file.
        encoding (str): Use this encoding on all the output streams. Default to
            the system's default encoding. Only applicable if is_save True.
        append (bool): If stdout/stderr refers to a file path, append to file
            rather than overwrite if it exists.
        results (list[SubprocessResult]): List of results. Populated with the
            result of each run instruction in `cmd`. Only when is_save is True.
    """

    def __init__(self,
                 cmd,
                 is_shell=False,
                 cwd=None,
                 is_save=False,
                 is_text=True,
                 stdout=None,
                 stderr=None,
                 encoding=None,
                 append=False):
        """Initialize the Cmd."""
        self.cmd = cmd
        self.is_shell = is_shell
        self.cwd = cwd
        self.is_save = is_save
        self.is_text = is_text if is_save else False

        if is_save:
            if stdout or stderr:
                raise ContextError(
                    "You can't set `stdout` or `stderr` when `save` is True.")
        self.stdout = stdout
        self.stderr = stderr
        self.encoding = encoding if encoding else config.default_cmd_encoding
        self.append = append

        self.results: list[SubprocessResult] = []

    @contextmanager
    def output_handles(self):
        """Return stdout + stderr output handles for subprocess.

        If self.stdout or self.stderr are None, will return None for whichever
        is None. This will result in the subprocess using the parent's handler
        for that output device.

        If self.stdout or self.stderr refers to a path, will open the file(s)
        at the given location for writing. Will create parent directories if
        these do not exist. self.append controls whether to overwrite or append
        files. Set self.encoding to write in text mode in the given encoding,
        otherwise the file handle opens in binary mode.

        The context manager will close the file handles (if any) when it exits.

        If self.stdout or self.stderr are the special values '/dev/null' or
        '/dev/stdout', will return the appropriate subprocess constants for
        these.

        Returns:
            (stdout, stderr): stdout & stderr handles to use when executing
                this command as a subprocess.
        """
        stdout = self.stdout
        stderr = self.stderr

        # fast short-circuit when no processing necessary
        if not stdout and not stderr:
            logger.debug("stdout & stderr inheriting from parent process.")
            yield stdout, stderr
        else:
            # append or overwrite
            mode = 'ab' if self.append else 'wb'
            files = []

            try:
                # stdout
                if stdout == '/dev/null':
                    logger.debug("redirecting stdout to /dev/null")
                    stdout = subprocess.DEVNULL
                elif stdout:
                    # by elimination, it's a path
                    logger.debug("writing stdout to %s in mode '%s'",
                                 stdout, mode)
                    Path(stdout).parent.mkdir(parents=True, exist_ok=True)
                    stdout = open(stdout, mode=mode)
                    files.append(stdout)

                #  stderr
                if stderr == '/dev/null':
                    logger.debug("redirecting stderr to /dev/null")
                    stderr = subprocess.DEVNULL
                elif stderr == '/dev/stdout':
                    logger.debug("redirecting stderr to /dev/stdout")
                    stderr = subprocess.STDOUT
                elif stderr:
                    # by elimination, it's a path
                    logger.debug("writing stderr to %s in mode '%s'",
                                 stderr, mode)
                    Path(stderr).parent.mkdir(parents=True, exist_ok=True)
                    stderr = open(stderr, mode=mode)
                    files.append(stderr)

                yield stdout, stderr
            finally:
                for f in files:
                    logger.debug("closing cmd file output %s", f.name)
                    f.close()

    def run(self):
        """Run the command as a subprocess.

        Raises:
            pypyr.errors.ContextError: The cmd executable instruction is
                formatted incorrectly.
            subprocess.CalledProcessError: Error executing the subprocess.
            FileNotFoundError: The system could not find the executable. It
                might not be on $PATH, or the path to the executable might be
                wrong.
        """
        cmd = self.cmd

        # this here because all outputs for this cmd write to the same
        # device/file handles - whether single cmd or list of commands.
        with self.output_handles() as (stdout, stderr):
            if isinstance(cmd, SimpleCommandTypes):
                # just one command, run as is
                self._run(cmd, stdout, stderr)
            elif isinstance(cmd, Sequence):
                # str explicitly checked before in if, so Sequence won't be str
                # except weirdness if passing bytearray (ByteString), but why
                # would you?
                for c in cmd:
                    # run each command in list of commands
                    self._run(c, stdout, stderr)
            else:
                raise ContextError(
                    f"{cmd} cmd should be either a simple string:\n"
                    "cmd: my-executable --arg\n\n"
                    "Or in the expanded syntax, set `run` to a string:\n\n"
                    "cmd:\n"
                    "  run: my-executable --arg\n"
                    "  cwd: ./mydir\n\n"
                    "Or set run to a list of commands:\n"
                    "cmd:\n"
                    "  run:\n"
                    "    - my-executable --arg\n"
                    "    - another-executable --arg2\n"
                    "  cwd: ../mydir/subdir")

    def _run(self, cmd, stdout=None, stderr=None):
        """Spawn subprocess for cmd with stdout + stderr output streams.

        If self.save is True, will append the completed process results to
        self.results, for both success and failure. Will strip trailing
        whitespace from output.

        Thereafter will raise an exception if anything had gone wrong with the
        subprocess.

        Args:
            cmd (str): The executable + args to spawn subprocess.
            stdout: Write stdout here. This can be None (attach to parent
                    process), an open file handle, or subprocess.DEVNULL.
            stderr: Write stderr here. This can be None (attach to parent
                    process), an open file handle, or subprocess.DEVNULL, or
                    subprocess.STDOUT to redirect to stdout.

        Raises:
            subprocess.CalledProcessError: Error executing the subprocess.
            FileNotFoundError: The system could not find the executable. It
                might not be on $PATH, or the path to the executable might be
                wrong.
        """
        if self.cwd:
            logger.debug("Processing command string in dir %s: %s",
                         self.cwd, self.cmd)
        else:
            logger.debug("Processing command string: %s", cmd)

        # why? If shell is True, python docs recommend passing args as a string
        # rather than as a sequence.
        # But not on windows, because windows API wants strings, not sequences.
        args = cmd if (
            self.is_shell or config.is_windows) else shlex.split(cmd)

        if self.is_save:
            # errs from _inside_ the subprocess will go to stderr and raise
            # via check_returncode. errs finding the executable will raise
            # right here - i.e won't end up in cmdOut.
            completed_process = subprocess.run(args,
                                               capture_output=True,
                                               cwd=self.cwd,
                                               encoding=self.encoding,
                                               shell=self.is_shell,
                                               text=self.is_text)

            stdout = completed_process.stdout
            stderr = completed_process.stderr

            if self.is_text:
                if stdout:
                    stdout = stdout.rstrip()
                if stderr:
                    stderr = stderr.rstrip()

            out = SubprocessResult(cmd=args,
                                   returncode=completed_process.returncode,
                                   stdout=stdout,
                                   stderr=stderr)

            # when capture is true, output doesn't write to stdout
            logger.info("stdout: %s", stdout)
            if stderr:
                logger.error("stderr: %s", stderr)

            # save output before checking/raising err that kicks out of method
            self.results.append(out)
            # don't swallow the error, because it's the Step swallow decorator
            # responsibility to decide to ignore or not.
            completed_process.check_returncode()
        else:
            # check=True throws CalledProcessError if exit code != 0
            subprocess.run(args,
                           check=True,
                           cwd=self.cwd,
                           shell=self.is_shell,
                           stdout=stdout,
                           stderr=stderr)

    def __eq__(self, other):
        """Check equality for all attributes."""
        if self is other:
            return True

        if type(other) is Command:
            return self.__dict__ == other.__dict__

        return NotImplemented


class SubprocessResult():
    """Result from a subprocess invocation.

    Attributes:
      cmd: The list or str args passed to the subprocess run instruction.
      returncode: The exit code of the process, negative for signals.
      stdout: The standard output (None if not captured).
      stderr: The standard error (None if not captured).

    Will also getitem in a dict-like way like r['returncode'] where dict is:
    {
        'returncode': proc.returncode,
        'stdout': stdout,
        'stderr': stderr
    }

    The dict-like accessor is only there for backwards compatibility, don't
    use it for anything new.
    """

    def __init__(self, cmd, returncode, stdout=None, stderr=None):
        """Initialize result."""
        self.cmd = cmd
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr

    def __getitem__(self, key):
        """Allow dict-like r['returncode'] access for backwards compat."""
        return self.__dict__[key]

    def __repr__(self) -> str:
        """Return repr."""
        args = ['cmd={!r}'.format(self.cmd),
                'returncode={!r}'.format(self.returncode)]
        if self.stdout is not None:
            args.append('stdout={!r}'.format(self.stdout))
        if self.stderr is not None:
            args.append('stderr={!r}'.format(self.stderr))
        return "{}({})".format(type(self).__name__, ', '.join(args))

    def __str__(self) -> str:
        """Get user friendly string."""
        return f"""\
cmd: {self.cmd}
returncode: {self.returncode}
stdout: {self.stdout}
stderr: {self.stderr}
"""

    def check_returncode(self) -> SubprocessError | None:
        """Return SubprocessError if the exit code is non-zero.

        This does not raise the error, it just returns it.

        Returns:
            None if no return code 0.
            SubprocessError if return code != 0.
        """
        if self.returncode:
            return SubprocessError(returncode=self.returncode,
                                   cmd=self.cmd,
                                   stdout=self.stdout,
                                   stderr=self.stderr)
        return None



"""pypyr step yaml for async subprocess commands - domain specific language."""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
from collections.abc import Mapping, Sequence
import logging

from pypyr.aio.subproc import Command, Commands
from pypyr.context import Context
from pypyr.errors import (ContextError,
                          KeyInContextHasNoValueError,
                          KeyNotInContextError)
import pypyr.utils.types
from pypyr.subproc import SimpleCommandTypes

logger = logging.getLogger(__name__)

AsyncCommandTypes = SimpleCommandTypes + (Sequence,)


class AsyncCmdStep():
    """A pypyr step to run executables/commands concurrently as a subprocess.

    This models a step that takes config like this in simple syntax:
        cmds:
            - <<cmd string 1>>
            - <<cmd string 2>>

    All the commands will run concurrently, in parallel.

    OR, expanded syntax is as a dict
        cmds:
            run: list[str | list[str]]. mandatory. command + args to execute.
                If list entry is another list[str], the sub-list will run in
                serial.
            save: bool. defaults False. save output to cmdOut. Treats output
                as text in the system's encoding and removes newlines at end.
            cwd: str/Pathlike. optional. Working directory for these commands.
            bytes (bool): Default False. When `save` return output bytes from
                cmds unaltered, without applying any encoding & text newline
                processing.
            encoding (str): Default None. When `save`, decode output with
                this encoding. The default of None uses the system encoding and
                should "just work".
            stdout (str | Path): Default None. Write stdout to this file path.
                Special value `/dev/null` writes to the system null device.
            stderr (str | Path): Default None. Write stderr to this file path.
                Special value `/dev/null` writes to the system null device.
                Special value `/dev/stdout` redirects err output to stdout.
            append (bool): Default False. When stdout/stderr a file, append
                rather than overwrite. Default is to overwrite.

    In expanded syntax, `run` can be a simple string or a list:
        cmds:
          run:
            - ./my-executable --arg
            - [./another-executable --arg, ./arb-executable arghere]
          save: False
          cwd: ./path/here

    As a list in simplified syntax:
        cmds:
          - my-executable --arg
          - ./another-executable --arg

    Any or all of the list items can use expanded syntax:
        cmds:
          - ./simple-cmd-here --arg1 value
          - run: cmd here
            save: False cwd: ./path/here
          - run:
              - my-executable --arg
              - ./another-executable --arg
            save: True
            cwd: ./path/here

    Any of the list items can in turn be a list. A sub-list will run in serial.

    In this example A, B.1 & C will start concurrently. B.2 will only run once
    B.1 is finished.

        cmds:
            - A
            - [B.1, B.2]
            - C

    If save is True, will save the output to context as cmdOut.

    cmdOut will be a list of pypyr.subproc.SubprocessResult objects, in order
    executed.

    SubprocessResult has the following properties:
    cmd: the cmd/args executed
    returncode: 0
    stdout: 'stdout str here. None if empty.'
    stderr: 'stderr str here. None if empty.'

    cmdOut.returncode is the exit status of the called process. Typically 0
    means OK. A negative value -N indicates that the child was terminated by
    signal N (POSIX only).

    The run_step method does the actual work. init parses the input yaml.

    Attributes:
        logger (logger): Logger instantiated by name of calling step.
        context: (pypyr.context.Context): The current pypyr Context.
        commands (pypyr.subproc.Commands): Commands to run as subprocess.
        is_shell (bool): True if subprocess should run through default shell.
        name (str): Name of calling step. Used for logging output & error
            messages.
    """

    def __init__(self,
                 name: str,
                 context: Context,
                 is_shell: bool = False) -> None:
        """Initialize the CmdStep.

        The step config in the context dict in simplified syntax:
            cmd: <<cmd string>>

        OR, as a dict in expanded syntax:
            cmd:
                run: str. mandatory. command + args to execute.
                save: bool. optional. defaults False. save output to cmdOut.
                cwd: str/path. optional. if specified, change the working
                     directory just for the duration of the command.

        `run` can be a single string, or it can be a list of string if there
        are multiple commands to execute with the same settings.

        OR, as a list:
            cmd:
                - my-executable --arg
                - ./another-executable --arg

        Any or all of the list items can be in expanded syntax.

        Args:
            name (str): Unique name for step. Likely __name__ of calling step.
            context (pypyr.context.Context): Look for step config in this
                context instance.
            is_shell (bool): Set to true to execute cmd through the default
                shell.
        """
        assert name, ("name parameter must exist for CmdStep.")
        assert context, ("context param must exist for CmdStep.")
        # this way, logs output as the calling step, which makes more sense
        # to end-user than a mystery steps.dsl.blah logging output.
        self.name = name
        self.logger = logging.getLogger(name)

        context.assert_key_has_value(key='cmds', caller=name)

        self.context = context
        self.is_shell = is_shell
        cmd_config = context.get_formatted('cmds')

        commands = Commands()
        if isinstance(cmd_config, SimpleCommandTypes):
            commands.append(Command(cmd_config, is_shell=is_shell))
        elif isinstance(cmd_config, Mapping):
            commands.append(self.create_command(cmd_config))
        elif isinstance(cmd_config, Sequence):
            for cmd in cmd_config:
                if isinstance(cmd, SimpleCommandTypes):
                    commands.append(Command(cmd, is_shell=is_shell))
                elif isinstance(cmd, Sequence):
                    commands.append(Command([cmd], is_shell=is_shell))
                elif isinstance(cmd, Mapping):
                    commands.append(self.create_command(cmd))
                else:
                    raise ContextError(
                        f"{cmd} in {name} cmds config is wrong.\n"
                        "Each list item should be either a simple string, or "
                        "a list to run in serial,\n"
                        "or a dict for expanded syntax:\n"
                        "cmds:\n"
                        "  - ./my-executable --arg\n"
                        "  - run:\n"
                        "      - ./another-executable --arg value\n"
                        "      - ./another-executable --arg value2\n"
                        "    cwd: ../mydir/subdir\n"
                        "  - run:\n"
                        "      - ./arb-executable1 --arg value1\n"
                        "      - [./arb-executable2.1, ./arb-executable2.2]\n"
                        "    cwd: ../mydir/arbdir\n"
                        "  - [./arb-executable3.1, ./arb-executable3.2]"
                    )
        else:
            raise ContextError(f"{name} cmds config should be either a list:\n"
                               "cmds:\n"
                               "  - ./my-executable --arg\n"
                               "  - subdir/executable --arg1\n\n"
                               "or a dictionary with a `run` sub-key:\n"
                               "cmds:\n"
                               "  run:\n"
                               "    - ./my-executable --arg\n"
                               "    - subdir/executable --arg1\n"
                               "  cwd: ./mydir\n\n"
                               "Any of the list items in root can be in "
                               "expanded syntax:\n"
                               "cmds:\n"
                               "  - ./my-executable --arg\n"
                               "  - subdir/executable --arg1\n"
                               "  - run:\n"
                               "      - ./arb-executable1 --arg value1\n"
                               "      - [./arb-executable2.1, "
                               "./arb-executable2.2]\n"
                               "    cwd: ../mydir/subdir\n"
                               "  - [./arb-executable3.1, ./arb-executable3.2]"
                               )

        self.commands: Commands = commands

    def create_command(self, cmd_input: Mapping) -> Command:
        """Create pypyr.aio.subproc.Command object from expanded step input."""
        try:
            cmd = cmd_input['run']  # can be str or list
            if not cmd:
                raise KeyInContextHasNoValueError(
                    f"cmds.run must have a value for {self.name}.\n"
                    "The `run` input should look something like this:\n"
                    "cmds:\n"
                    "  run:\n"
                    "    - ./arb-executable1 --arg value1\n"
                    "    - ./arb-executable2 --arg value2\n"
                    "  cwd: ../mydir/arbdir")
        except KeyError as err:
            raise KeyNotInContextError(
                f"cmds.run doesn't exist for {self.name}.\n"
                "The input should look like this in expanded syntax:\n"
                "cmds:\n"
                "  run:\n"
                "    - ./my-executable --arg\n"
                "    - subdir/executable --arg1\n"
                "  cwd: ./mydir\n\n"
                "If you're passing in a list of commands, each command should "
                "be a simple string,\n"
                "or a sub-list of commands to run in serial,\n"
                "or a dict with a `run` entry:\n"
                "cmds:\n"
                "  - ./my-executable --arg\n"
                "  - run: ./another-executable --arg value\n"
                "    cwd: ../mydir/subdir\n"
                "  - run:\n"
                "      - ./arb-executable1 --arg value1\n"
                "      - [./arb-executable2.1, ./arb-executable2.2]\n"
                "    cwd: ../mydir/arbdir\n"
                "  - [./arb-executable3.1, ./arb-executable3.2]"
            ) from err

        is_save = pypyr.utils.types.cast_to_bool(cmd_input.get('save', False))

        cwd = cmd_input.get('cwd')

        is_bytes = cmd_input.get('bytes')
        is_text = not is_bytes if is_save else False

        stdout = cmd_input.get('stdout')
        stderr = cmd_input.get('stderr')

        if is_save:
            if stderr or stderr:
                raise ContextError(
                    "You can't set `stdout` or `stderr` when `save` is True.")

        encoding = cmd_input.get('encoding')
        append = cmd_input.get('append', False)
        is_shell_override = cmd_input.get('shell', None)

        is_shell = (
            self.is_shell if is_shell_override is None else is_shell_override)

        return Command(cmd=cmd,
                       is_shell=is_shell,
                       cwd=cwd,
                       is_save=is_save,
                       is_text=is_text,
                       stdout=stdout,
                       stderr=stderr,
                       encoding=encoding,
                       append=append)

    def run_step(self) -> None:
        """Spawn subprocesses to run the commands asynchronously.

        If cmd.is_save==True, save aggregate result of all commands to context
        'cmdOut'.

        cmdOut will be a list of pypyr.subproc.SubprocessResult or Exception
        objects, in order executed.

        SubprocessResult has the following properties:
        cmd: the cmd/args executed
        returncode: 0
        stdout: 'stdout str here. None if empty.'
        stderr: 'stderr str here. None if empty.'
        """
        try:
            self.commands.run()
        finally:
            if self.commands.is_save:
                self.logger.debug("saving results to cmdOut")
                self.context['cmdOut'] = self.commands.results
            else:
                self.logger.debug(
                    "save is False: not saving results to cmdOut")



"""version.py unit tests."""
import pypyr.version
import platform


def test_get_version():
    """Test version as expected."""
    actual = pypyr.version.get_version()
    expected = (f'pypyr {pypyr.version.__version__} '
                f'python {platform.python_version()}')
    assert actual == expected, "version not returning correctly"



"""env.py unit tests."""
import os
import pytest
from unittest.mock import patch, DEFAULT
from pypyr.context import Context
from pypyr.errors import KeyNotInContextError
import pypyr.steps.env

# ------------------------- env base -----------------------------------------#


def test_env_throws_on_empty_context():
    """Context must exist."""
    with pytest.raises(AssertionError) as err_info:
        pypyr.steps.env.run_step(None)

    assert str(err_info.value) == ("context must have value "
                                   "for pypyr.steps.env")


def test_env_throws_if_all_env_context_missing():
    """Env context keys must exist."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.env.run_step(Context({'arbkey': 'arbvalue'}))

    assert str(err_info.value) == ("context['env'] doesn't exist. It must "
                                   "exist for pypyr.steps.env.")


def test_env_throws_if_env_context_wrong_type():
    """Env context keys must exist."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.env.run_step(
            Context({'env': {'envSet': 'it shouldnt be a string'}}))

    assert str(err_info.value) == ("context must contain "
                                   "any combination of env['get'], env['set'] "
                                   "or env['unset'] for pypyr.steps.env")


def test_env_only_calls_get():
    """Call only get."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'env': {'get': {
            'key2': 'ARB_GET_ME1',
            'key4': 'ARB_GET_ME2'
        }}
    })

    with patch.multiple('pypyr.steps.env',
                        env_get=DEFAULT,
                        env_set=DEFAULT,
                        env_unset=DEFAULT
                        ) as mock_env:
        pypyr.steps.env.run_step(context)

    mock_env['env_get'].assert_called_once()
    mock_env['env_set'].assert_called_once()
    mock_env['env_unset'].assert_called_once()


def test_env_only_calls_set():
    """Call only set."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'env': {'set': {
            'ARB_SET_ME1': 'key2',
            'ARB_SET_ME2': 'key1'
        }}
    })

    with patch.multiple('pypyr.steps.env',
                        env_get=DEFAULT,
                        env_set=DEFAULT,
                        env_unset=DEFAULT
                        ) as mock_env:
        pypyr.steps.env.run_step(context)

    mock_env['env_get'].assert_called_once()
    mock_env['env_set'].assert_called_once()
    mock_env['env_unset'].assert_called_once()


def test_env_only_calls_unset():
    """Call only unset."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'env': {'unset': [
            'ARB_DELETE_ME1',
            'ARB_DELETE_ME2'
        ]}
    })

    with patch.multiple('pypyr.steps.env',
                        env_get=DEFAULT,
                        env_set=DEFAULT,
                        env_unset=DEFAULT
                        ) as mock_env:
        pypyr.steps.env.run_step(context)

    mock_env['env_get'].assert_called_once()
    mock_env['env_set'].assert_called_once()
    mock_env['env_unset'].assert_called_once()


def test_env_all_operations():
    """Env should run all specified operations."""
    os.environ['ARB_GET_ME1'] = 'arb value from $ENV ARB_GET_ME1'
    os.environ['ARB_GET_ME2'] = 'arb value from $ENV ARB_GET_ME2'
    os.environ['ARB_DELETE_ME1'] = 'arb value from $ENV ARB_DELETE_ME1'
    os.environ['ARB_DELETE_ME2'] = 'arb value from $ENV ARB_DELETE_ME2'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'env': {
                'get': {
                    'key2': 'ARB_GET_ME1',
                    'key4': 'ARB_GET_ME2'},
                'set': {
                    'ARB_SET_ME1': 'value 4',
                    'ARB_SET_ME2': 'go go {key2} end end'},
                'unset': [
                    'ARB_DELETE_ME1',
                    'ARB_DELETE_ME2']}
    })

    pypyr.steps.env.run_step(context)

    assert context['key1'] == 'value1'
    assert context['key2'] == 'arb value from $ENV ARB_GET_ME1'
    assert context['key3'] == 'value3'
    assert context['key4'] == 'arb value from $ENV ARB_GET_ME2'
    assert os.environ['ARB_SET_ME1'] == 'value 4'
    assert os.environ['ARB_SET_ME2'] == ('go go arb value from '
                                         '$ENV ARB_GET_ME1 end end')
    assert 'ARB_DELETE_ME1' not in os.environ
    assert 'ARB_DELETE_ME2' not in os.environ

    del os.environ['ARB_GET_ME1']
    del os.environ['ARB_GET_ME2']
    del os.environ['ARB_SET_ME1']
    del os.environ['ARB_SET_ME2']

# ------------------------- env base -----------------------------------------#
#
# ------------------------- envGet -------------------------------------------#


def test_envget_pass():
    """Env get success case."""
    os.environ['ARB_DELETE_ME1'] = 'arb value from $ENV ARB_DELETE_ME1'
    os.environ['ARB_DELETE_ME2'] = 'arb value {from} $ENV ARB_DELETE_ME2'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'keyname': 'key',
        'delete_substitute': 'DELETE',
        'env': {'get': {
            'key2': 'ARB_DELETE_ME1',
            '{keyname}4': 'ARB_{delete_substitute}_ME2'
        }}
    })

    pypyr.steps.env.env_get(context)

    assert context['key1'] == 'value1'
    assert context['key2'] == 'arb value from $ENV ARB_DELETE_ME1'
    assert context['key3'] == 'value3'
    assert context['key4'] == 'arb value {from} $ENV ARB_DELETE_ME2'

    del os.environ['ARB_DELETE_ME1']
    del os.environ['ARB_DELETE_ME2']


def test_envget_env_doesnt_exist():
    """Env get when $ENV doesn't exist."""
    os.environ['ARB_DELETE_ME2'] = 'arb from pypyr context ARB_DELETE_ME2'
    try:
        del os.environ['ARB_DELETE_ME1']
    except KeyError:
        pass

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'env': {'get': {
            'key2': 'ARB_DELETE_ME1',
            'key4': 'ARB_DELETE_ME2'
        }}
    })

    with pytest.raises(KeyError) as err_info:
        pypyr.steps.env.env_get(context)

    assert str(err_info.value) == "'ARB_DELETE_ME1'"
    del os.environ['ARB_DELETE_ME2']


def test_envget_returns_false():
    """Env get returns false if env>get not specified."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'env': {'Xget': {
            'key2': 'ARB_DELETE_ME1'
        }}
    })

    assert not pypyr.steps.env.env_get(context)
# ------------------------- envGet -------------------------------------------#

# ------------------------- envSet -------------------------------------------#


def test_envset_pass():
    """Env set success case."""
    # Deliberately have 1 pre-existing $ENV to update, and 1 unset so can
    # create it anew as part of test
    os.environ['ARB_DELETE_ME2'] = 'arb from pypyr context ARB_DELETE_ME2'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'env': {'set': {
            'ARB_DELETE_ME1': 'text value 2',
            'ARB_DELETE_ME2': 'text value 1'
        }}
    })

    assert pypyr.steps.env.env_set(context)

    assert os.environ['ARB_DELETE_ME1'] == 'text value 2'
    assert os.environ['ARB_DELETE_ME2'] == 'text value 1'

    del os.environ['ARB_DELETE_ME1']
    del os.environ['ARB_DELETE_ME2']


def test_envset_with_string_interpolation():
    """Env set success case."""
    # Deliberately have 1 pre-existing $ENV to update, and 1 unset so can
    # create it anew as part of test
    os.environ['ARB_DELETE_ME2'] = 'arb from pypyr context ARB_DELETE_ME2'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'del_substitute': 'DELETE',
        'env': {'set': {
            'ARB_DELETE_ME1': 'blah blah {key2} and {key1} goes here.',
            'ARB_DELETE_ME2': 'plain old string',
            'ARB_{del_substitute}_ME3': '{key3}'
        }}
    })

    assert pypyr.steps.env.env_set(context)

    assert os.environ['ARB_DELETE_ME1'] == ('blah blah value2 and value1 goes '
                                            'here.')
    assert os.environ['ARB_DELETE_ME2'] == 'plain old string'
    assert os.environ['ARB_DELETE_ME3'] == 'value3'

    del os.environ['ARB_DELETE_ME1']
    del os.environ['ARB_DELETE_ME2']
    del os.environ['ARB_DELETE_ME3']


def test_envset_returns_false():
    """Env set returns false if env>set doesn't exist.."""
    # Deliberately have 1 pre-existing $ENV to update, and 1 unset so can
    # create it anew as part of test
    os.environ['ARB_DELETE_ME2'] = 'arb from pypyr context ARB_DELETE_ME2'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'env': {'blah': {
            'ARB_DELETE_ME1': 'blah blah {key2} and {key1} goes here.'
        }}
    })

    assert not pypyr.steps.env.env_set(context)

# ------------------------- envSet -------------------------------------------#

# ------------------------- envUnset -----------------------------------------#


def test_envunset_pass():
    """Env unset success case."""
    # Deliberately have 1 pre-existing $ENV to update, and 1 unset so can
    # create it anew as part of test
    os.environ['ARB_DELETE_ME1'] = 'arb from pypyr context ARB_DELETE_ME1'
    os.environ['ARB_DELETE_ME2'] = 'arb from pypyr context ARB_DELETE_ME2'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'env': {'unset': [
            'ARB_DELETE_ME1',
            'ARB_DELETE_ME2'
        ]}
    })

    assert pypyr.steps.env.env_unset(context)

    assert 'ARB_DELETE_ME1' not in os.environ
    assert 'ARB_DELETE_ME2' not in os.environ


def test_envunset_pass_wuth_interpolation():
    """Env unset success case with interpolation."""
    # Deliberately have 1 pre-existing $ENV to update, and 1 unset so can
    # create it anew as part of test
    os.environ['ARB_DELETE_ME1'] = 'arb from pypyr context ARB_DELETE_ME1'
    os.environ['ARB_DELETE_ME2'] = 'arb from pypyr context ARB_DELETE_ME2'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        "one": 1,
        'del': 'DELETE',
        'del_full': 'ARB_DELETE_ME2',
        'env': {'unset': [
            'ARB_{del}_ME{one}',
            '{del_full}'
        ]}
    })

    assert pypyr.steps.env.env_unset(context)

    assert 'ARB_DELETE_ME1' not in os.environ
    assert 'ARB_DELETE_ME2' not in os.environ


def test_envunset_doesnt_exist():
    """Env unset success case where env doesn't exist."""
    # Make sure $ENV isn't set
    try:
        del os.environ['ARB_DELETE_SNARK']
    except KeyError:
        pass

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'env': {'unset': [
                'ARB_DELETE_SNARK'
                ]}
    })

    assert pypyr.steps.env.env_unset(context)

    assert 'ARB_DELETE_SNARK' not in os.environ


def test_envunset_returns_false():
    """Env unset returns false if not env>unset not specified.."""
    context = Context({
        'env': {'set': {
            'ARB_DELETE_ME1': 'blah blah {key2} and {key1} goes here.'
        }}
    })

    assert not pypyr.steps.env.env_unset(context)
# ------------------------- envUnset -----------------------------------------#



"""backoffcache.py unit tests."""
import pytest
from unittest.mock import patch
from pypyr.errors import PyModuleNotFoundError
import pypyr.cache.backoffcache as backoffcache
import pypyr.retries

# region load_backoff_callable


def test_load_backoff_callable_module_not_found():
    """Module not found raises."""
    with pytest.raises(PyModuleNotFoundError):
        backoffcache.load_backoff_callable('blah-blah-xyz-argh.unlikely-name')


def test_load_backoff_callable_attr_not_found():
    """Attribute not found raises."""
    with pytest.raises(AttributeError) as err:
        backoffcache.load_backoff_callable(
            'tests.arbpack.arbcallables.IDontExist')

    assert str(err.value) == (
        "module 'tests.arbpack.arbcallables' has no attribute 'IDontExist'")


def test_load_backoff_callable_absolute():
    """Load backoff callable with absolute name: package.module.class."""
    f = backoffcache.load_backoff_callable(
        'tests.arbpack.arbcallables.ArbCallable')

    callable_ref = f('ctor in')
    assert callable_ref('arg in 1') == 'from callable: ctor in arg in 1'
    assert callable_ref('arg in 2') == 'from callable: ctor in arg in 2'


def test_load_backoff_callable_bare():
    """Load backoff callable with bare name (no dot)."""
    with pytest.raises(ValueError) as err:
        backoffcache.load_backoff_callable('local_test_arb_callable')

    assert str(err.value) == (
        "Trying to find back-off strategy 'local_test_arb_callable'. If this "
        "is a built-in back-off strategy, are you sure you got the name right?"
        "\nIf you're trying to load a custom callable, name should be in "
        "format 'package.module.ClassName' or 'mod.ClassName'.")

# endregion load_backoff_callable

# region BackoffCache

# region BackoffCache: get_backoff


def test_get_backoff():
    """Backoff callable loads ok."""
    with patch('pypyr.cache.backoffcache.load_backoff_callable') as mock:
        mock.return_value = lambda x: f"{x}test"
        f = backoffcache.BackoffCache().get_backoff("arb")

    mock.assert_called_once_with('arb')
    assert f("arb") == "arbtest"

    # it made a copy
    assert 'arb' not in pypyr.retries.builtin_backoffs


def test_get_backoff_builtin():
    """Load built-in backoff callable."""
    f = backoffcache.BackoffCache().get_backoff('fixed')

    callable_ref = f(sleep=123, max_sleep=456)
    assert callable_ref(789) == 123
    assert callable_ref(999) == 123


@pytest.fixture
def no_cache(monkeypatch):
    """Set no cache."""
    monkeypatch.setattr('pypyr.cache.cache.config.no_cache', True)


def test_get_backoff_builtin_no_cache(no_cache):
    """Load built-in backoff callable when no_cache true."""
    backoff_cache = backoffcache.BackoffCache()
    f = backoff_cache.get_backoff('fixed')

    callable_ref = f(sleep=123, max_sleep=456)
    assert callable_ref(789) == 123
    assert callable_ref(999) == 123

    f2 = backoff_cache.get_backoff('fixed')
    callable_ref = f2(sleep=789, max_sleep=1000)
    assert callable_ref(111) == 789

    # didn't add anything to cache even when calling same built-in twice,
    # bypassing it entirely
    assert backoff_cache._cache == pypyr.retries.builtin_backoffs

# endregion BackoffCache: get_backoff

# region BackoffCache: clear


def test_clear_doesnt_wipe_builtins():
    """Clear shouldn't touch builtins."""
    # Also doesn't touch the og look-up dict.
    bc = backoffcache.BackoffCache()

    og_len = len(bc._cache)
    assert og_len > 1
    assert 'fixed' in bc._cache

    _ = bc.get_backoff('tests.arbpack.arbcallables.ArbCallable')
    assert len(bc._cache) == og_len + 1

    bc.clear()

    # didn't touch the class's built-in dict
    assert len(bc._cache) == og_len
    assert 'fixed' in bc._cache

    # twice in a row to verify clear also makes a copy
    _ = bc.get_backoff('tests.arbpack.arbcallables.ArbCallable')
    assert len(bc._cache) == og_len + 1
    assert len(pypyr.retries.builtin_backoffs) == og_len

    bc.clear()

    # didn't touch the class's built-in dict
    assert len(bc._cache) == og_len
    assert 'fixed' in bc._cache

    # didn't touch the shared built-in dict constant
    assert len(pypyr.retries.builtin_backoffs) == og_len
    assert 'fixed' in pypyr.retries.builtin_backoffs


# endregion BackoffCache: clear

# endregion BackoffCache



"""fileread.py unit tests."""
from unittest.mock import mock_open, patch

import pytest

from pypyr.context import Context
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
import pypyr.steps.fileread as fileread


def test_fileread_no_input_raises():
    """No input."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        fileread.run_step(context)

    assert str(err_info.value) == ("context['fileRead'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.fileread.")


def test_fileread_none_raises():
    """Input exists but is None."""
    context = Context({
        'k1': 'v1',
        'fileRead': None})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        fileread.run_step(context)

    assert str(err_info.value) == ("context['fileRead'] must have a "
                                   "value for pypyr.steps.fileread.")


def test_fileread_none_path_raises():
    """None path raises."""
    context = Context({
        'fileRead': {
            'path': None}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        fileread.run_step(context)

    assert str(err_info.value) == ("context['fileRead']['path'] must have a "
                                   "value for pypyr.steps.fileread.")


def test_fileread_empty_path_raises():
    """Empty path raises."""
    context = Context({
        'fileRead': {
            'path': ''}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        fileread.run_step(context)

    assert str(err_info.value) == ("context['fileRead']['path'] must have a "
                                   "value for pypyr.steps.fileread.")


def test_fileread_no_key_raises():
    """No key raises."""
    context = Context({
        'fileRead': {
            'path': '/arb'}})

    with pytest.raises(KeyNotInContextError) as err_info:
        fileread.run_step(context)

    assert str(err_info.value) == ("context['fileRead']['key'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.fileread.")


def test_fileread_none_key_raises():
    """None key raises."""
    context = Context({
        'fileRead': {
            'path': '/arb',
            'key': None}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        fileread.run_step(context)

    assert str(err_info.value) == ("context['fileRead']['key'] must have a "
                                   "value for pypyr.steps.fileread.")


def test_fileread_empty_key_raises():
    """Empty key raises."""
    context = Context({
        'fileRead': {
            'path': '/arb',
            'key': ''}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        fileread.run_step(context)

    assert str(err_info.value) == ("context['fileRead']['key'] must have a "
                                   "value for pypyr.steps.fileread.")

# endregion validation

# region text mode


def test_fileread_defaults():
    """Read file with minimal defaults."""
    context = Context({
        'fileRead': {
            'path': '/arb',
            'key': 'out'}})

    with patch('pypyr.steps.fileread.open',
               mock_open(read_data='one\ntwo\nthree')) as mocked_open:
        fileread.run_step(context)

    assert context['out'] == 'one\ntwo\nthree'

    mocked_open.assert_called_once_with('/arb', 'r', encoding=None)


def test_fileread_defaults_with_encoding(monkeypatch):
    """Read file with minimal defaults and encoding set by config."""
    monkeypatch.setattr('pypyr.config.config.default_encoding', 'arbe')
    context = Context({
        'fileRead': {
            'path': '/arb',
            'key': 'out'}})

    with patch('pypyr.steps.fileread.open',
               mock_open(read_data='one\ntwo\nthree')) as mocked_open:
        fileread.run_step(context)

    assert context['out'] == 'one\ntwo\nthree'

    mocked_open.assert_called_once_with('/arb', 'r', encoding='arbe')


def test_fileread_all():
    """Read file with all args."""
    context = Context({
        'fileRead': {
            'path': '/arb',
            'key': 'out',
            'encoding': 'arb'}})

    with patch('pypyr.steps.fileread.open',
               mock_open(read_data='one\ntwo\nthree')) as mocked_open:
        fileread.run_step(context)

    assert context['out'] == 'one\ntwo\nthree'

    mocked_open.assert_called_once_with('/arb', 'r', encoding='arb')


def test_fileread_all_encoding_override(monkeypatch):
    """Read file with all args, overriding config encoding."""
    monkeypatch.setattr('pypyr.config.config.default_encoding', 'XXX')
    context = Context({
        'fileRead': {
            'path': '/arb',
            'key': 'out',
            'encoding': 'arb'}})

    with patch('pypyr.steps.fileread.open',
               mock_open(read_data='one\ntwo\nthree')) as mocked_open:
        fileread.run_step(context)

    assert context['out'] == 'one\ntwo\nthree'

    mocked_open.assert_called_once_with('/arb', 'r', encoding='arb')

# endregion text mode

# region binary mode


def test_fileread_binary_true():
    """Read file with binary true."""
    context = Context({
        'fileRead': {
            'path': '/arb',
            'key': 'out',
            'binary': True}})

    with patch('pypyr.steps.fileread.open',
               mock_open(read_data=b'12345')) as mocked_open:
        fileread.run_step(context)

    assert context['out'] == b'12345'

    mocked_open.assert_called_once_with('/arb', 'rb', encoding=None)


def test_fileread_binary_explicit_false():
    """Read file with binary explicit false."""
    context = Context({
        'fileRead': {
            'path': '/arb',
            'key': 'out',
            'binary': False}})

    with patch('pypyr.steps.fileread.open',
               mock_open(read_data='one\ntwo\nthree')) as mocked_open:
        fileread.run_step(context)

    assert context['out'] == 'one\ntwo\nthree'

    mocked_open.assert_called_once_with('/arb', 'r', encoding=None)

# endregion binary mode

# region substitutions


def test_fileread_substitutions():
    """Read file with substitutions."""
    context = Context({
        'p': '/arb',
        'k': 'out',
        'b': False,
        'c': 'arb',
        'fileRead': {
            'path': '{p}',
            'key': '{k}',
            'binary': '{b}',
            'encoding': '{c}'}})

    with patch('pypyr.steps.fileread.open',
               mock_open(read_data='one\ntwo\nthree')) as mocked_open:
        fileread.run_step(context)

    assert context['out'] == 'one\ntwo\nthree'

    assert context == {
        'p': '/arb',
        'k': 'out',
        'b': False,
        'c': 'arb',
        'out': 'one\ntwo\nthree',
        'fileRead': {
            'path': '{p}',
            'key': '{k}',
            'binary': '{b}',
            'encoding': '{c}'}}

    mocked_open.assert_called_once_with('/arb', 'r', encoding='arb')

# endregion substitutions



"""fetchtoml.py unit tests."""
from unittest.mock import mock_open, patch

import pytest

try:
    # can get rid of the try soon as py 3.11 min supported version
    # reason is py 3.11 includes tomli in stdlib
    from tomllib import TOMLDecodeError
except ModuleNotFoundError:
    from tomli import TOMLDecodeError

from pypyr.context import Context
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
import pypyr.steps.fetchtoml as tomlfetcher


def test_fetch_toml_no_path_raises():
    """None path raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        tomlfetcher.run_step(context)

    assert str(err_info.value) == ("context['fetchToml'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.fetchtoml.")


def test_fetch_toml_empty_path_raises():
    """Empty path raises."""
    context = Context({
        'fetchToml': {
            'path': None}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        tomlfetcher.run_step(context)

    assert str(err_info.value) == ("context['fetchToml']['path'] must have a "
                                   "value for pypyr.steps.fetchtoml.")


def test_fetch_toml_pass(fs):
    """Relative path to toml should succeed."""
    in_path = './tests/testfiles/test.toml'
    fs.create_file(in_path, contents="""key1 = "value1"
key2 = "value2"
key3 = "value3"
""")

    context = Context({
        'ok1': 'ov1',
        'fetchToml': {
            'path': in_path}})

    tomlfetcher.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 5, "context should have 5 items"
    assert context['ok1'] == 'ov1'
    assert 'fetchTomlPath' not in context
    assert context["key1"] == "value1", "key1 should be value2"
    assert context["key2"] == "value2", "key2 should be value2"
    assert context["key3"] == "value3", "key3 should be value2"


def test_fetch_toml_pass_with_string(fs):
    """Relative path to toml should succeed with string input."""
    in_path = './tests/testfiles/test.toml'
    fs.create_file(in_path, contents="""key1 = "value1"
key2 = "value2"
key3 = "value3"
""")

    context = Context({
        'ok1': 'ov1',
        'fetchToml': in_path})

    tomlfetcher.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 5, "context should have 5 items"
    assert context['ok1'] == 'ov1'
    assert 'fetchTomlPath' not in context
    assert context["key1"] == "value1", "key1 should be value2"
    assert context["key2"] == "value2", "key2 should be value2"
    assert context["key3"] == "value3", "key3 should be value2"


def test_fetch_toml_pass_with_path_substitution(fs):
    """Relative path to toml with string substitution on path."""
    in_path = './tests/testfiles/test.toml'
    fs.create_file(in_path, contents="""key1 = "value1"
key2 = "value2"
key3 = "value3"
""")

    context = Context({
        'ok1': 'ov1',
        'fileName': 'test',
        'fetchToml': {
            'path': './tests/testfiles/{fileName}.toml'}})

    tomlfetcher.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 6, "context should have 6 items"
    assert context['ok1'] == 'ov1'
    assert 'fetchTomlPath' not in context
    assert context["key1"] == "value1", "key1 should be value2"
    assert context["key2"] == "value2", "key2 should be value2"
    assert context["key3"] == "value3", "key3 should be value2"


def test_fetchtoml_with_destination():
    """Toml writes to destination key."""
    context = Context({
        'fetchToml': {
            'path': '/arb/arbfile',
            'key': 'outkey'}})

    toml_bytes = b'k1 = "v1"\nk2 = "v2"'
    with patch('pypyr.toml.open', mock_open(read_data=toml_bytes)):
        tomlfetcher.run_step(context)

    assert context['outkey'] == {'k1': 'v1', 'k2': 'v2'}
    assert len(context) == 2


def test_fetchtoml_with_destination_int():
    """Toml writes to destination key that's not a string."""
    context = Context({
        'fetchToml': {
            'path': '/arb/arbfile',
            'key': 99}})

    toml_bytes = b'k1 = "v1"\nk2 = "v2"'
    with patch('pypyr.toml.open', mock_open(read_data=toml_bytes)):
        tomlfetcher.run_step(context)

    assert context[99] == {'k1': 'v1', 'k2': 'v2'}
    assert len(context) == 2


def test_fetchtoml_with_destination_formatting():
    """Toml writes to destination key found by formatting expression."""
    context = Context({
        'keyhere': {'sub': ['outkey', 2, 3]},
        'fetchToml': {
            'path': '/arb/arbfile',
            'key': '{keyhere[sub][0]}'}})

    toml_bytes = b'1 = 2\n2 = 3'
    with patch('pypyr.toml.open', mock_open(read_data=toml_bytes)):
        tomlfetcher.run_step(context)

    assert len(context) == 3
    # toml spec always interprets keys as strings
    assert context['outkey'] == {'1': 2, '2': 3}
    assert context['keyhere'] == {'sub': ['outkey', 2, 3]}


def test_fetchtoml_invalid_toml_fails():
    """Invalid toml document fails to load."""
    context = Context({
        'ok1': 'ov1',
        'fetchToml': {
            'path': '/arb/arbfile'}})

    with patch('pypyr.toml.open', mock_open(read_data=b'[1,2,3]')):
        with pytest.raises(TOMLDecodeError):
            tomlfetcher.run_step(context)



"""parsercache.py unit tests."""
import pytest
from unittest.mock import patch
from pypyr.errors import PyModuleNotFoundError
import pypyr.cache.parsercache as parsercache

# ------------------------- load_the_parser --------------------------------#


def test_load_the_parser_module_not_found():
    """Module not found raises."""
    with pytest.raises(PyModuleNotFoundError):
        parsercache.load_the_parser('blah-blah-unlikely-name')


def test_load_the_parser_attr_not_found():
    """Attribute not found raises."""
    with pytest.raises(AttributeError):
        parsercache.load_the_parser(__name__)


def test_load_the_parser_ok():
    """Module with attribute returns function."""
    parsercache.contextparser_cache.clear()
    f = parsercache.load_the_parser('tests.arbpack.arbparser')
    response = f(context_arg='blah')

    assert response == {'parsed_context': 'blah'}
# ------------------------- END load_the_parser ----------------------------#

# ------------------------- ContextParserCache -----------------------------#

# ------------------------- ContextParserCache: get_context_parser ---------#


def test_get_pype_loader_specified():
    """Loader passes loader arg."""
    with patch('pypyr.cache.parsercache.load_the_parser') as mock:
        mock.return_value = lambda x: f"{x}test"
        f = parsercache.ContextParserCache().get_context_parser("arbparser")

    mock.assert_called_once_with('arbparser')
    assert f("arb") == "arbtest"
# ------------------------- ContextParserCache: get_context_parser ---------#

# ------------------------- END ContextParserCache --------------------------#



"""cache.py unit tests."""
import logging
from unittest.mock import call, MagicMock

import pytest

from pypyr.cache.cache import Cache
from tests.common.utils import patch_logger


@pytest.fixture
def no_cache(monkeypatch):
    """Set no cache."""
    monkeypatch.setattr('pypyr.cache.cache.config.no_cache', True)


def test_cache_get_miss():
    """Cache get should execute creator."""
    cache = Cache()
    creator_mock = MagicMock()
    creator_mock.return_value = "created obj"

    with patch_logger('pypyr.cache', logging.DEBUG) as mock_logger_debug:
        obj = cache.get('one', lambda: creator_mock("1"))

    assert obj == "created obj"
    creator_mock.assert_called_once_with("1")
    mock_logger_debug.assert_called_once_with(
        "`one` not found in cache. . . creating")


def test_cache_get_hit():
    """Cache get where found in cache shouldn't execute creator."""
    cache = Cache()
    creator_mock = MagicMock()
    creator_mock.side_effect = ["created obj1", "created obj2", "created obj3"]

    with patch_logger('pypyr.cache', logging.DEBUG) as mock_logger_debug:
        obj1 = cache.get('one', lambda: creator_mock("1"))
        obj2 = cache.get('one', lambda: creator_mock("2"))
        obj3 = cache.get('one', lambda: creator_mock("3"))

    assert obj1 == "created obj1"
    assert obj2 == "created obj1"
    assert obj3 == "created obj1"

    creator_mock.assert_called_once_with("1")

    assert mock_logger_debug.mock_calls == [
        call("`one` not found in cache. . . creating"),
        call("`one` loading from cache"),
        call("`one` loading from cache")]

    obj4 = creator_mock("4")
    assert obj4 == "created obj2"


def test_cache_get_hit_no_cache(no_cache):
    """Cache get with no_cache set should run creator each time."""
    cache = Cache()
    creator_mock = MagicMock()
    creator_mock.side_effect = ["created obj1", "created obj2", "created obj3",
                                "created obj4"]

    with patch_logger('pypyr.cache', logging.DEBUG) as mock_logger_debug:
        obj1 = cache.get('one', lambda: creator_mock("1"))
        obj2 = cache.get('one', lambda: creator_mock("2"))
        obj3 = cache.get('one', lambda: creator_mock("3"))

    assert obj1 == "created obj1"
    assert obj2 == "created obj2"
    assert obj3 == "created obj3"

    assert creator_mock.mock_calls == [
        call('1'),
        call('2'),
        call('3')
    ]

    assert mock_logger_debug.mock_calls == [
        call("no cache mode enabled. creating `one` sans cache"),
        call("no cache mode enabled. creating `one` sans cache"),
        call("no cache mode enabled. creating `one` sans cache")]

    obj4 = creator_mock("4")
    assert obj4 == "created obj4"


def test_cache_multiple_items_get_hit_closures():
    """Cache with multiple items work."""
    cache = Cache()

    def closure(x, y):
        def inner():
            return x + y
        return inner

    with patch_logger('pypyr.cache', logging.DEBUG) as mock_logger_debug:
        obj1 = cache.get('one', closure(2, 3))
        obj2 = cache.get('two', closure(4, 5))
        obj3 = cache.get('three', closure(6, 7))
        obj4 = cache.get('one', closure(8, 9))

    assert obj1 == 5
    assert obj2 == 9
    assert obj3 == 13
    assert obj4 == 5

    assert mock_logger_debug.mock_calls == [
        call("`one` not found in cache. . . creating"),
        call("`two` not found in cache. . . creating"),
        call("`three` not found in cache. . . creating"),
        call("`one` loading from cache")]


def test_cache_clear():
    """Cache should clear on clear()."""
    cache = Cache()

    def closure(x, y):
        def inner():
            return x + y
        return inner

    with patch_logger('pypyr.cache', logging.DEBUG) as mock_logger_debug:
        obj1 = cache.get('one', closure(2, 3))
        obj2 = cache.get('two', closure(4, 5))
        obj3 = cache.get('one', closure(6, 7))

    cache.clear()
    assert len(cache._cache) == 0
    assert obj1 == 5
    assert obj2 == 9
    assert obj3 == 5

    assert mock_logger_debug.mock_calls == [
        call("`one` not found in cache. . . creating"),
        call("`two` not found in cache. . . creating"),
        call("`one` loading from cache")]

    obj1 = cache.get('one', closure(2, 3))
    obj2 = cache.get('two', closure(4, 5))
    obj3 = cache.get('one', closure(6, 7))

    assert len(cache._cache) == 2
    assert obj1 == 5
    assert obj2 == 9
    assert obj3 == 5



"""Unit tests for subproc.py."""
import subprocess

import pytest

from pypyr.context import Context
from pypyr.dsl import PyString
from pypyr.errors import ContextError, SubprocessError
from pypyr.subproc import Command, SubprocessResult

# region Command


def test_subproc_command_eq():
    """Command instance does equality check on all members."""
    c = Command('cmd')
    assert c == c
    assert Command('cmd') == Command('cmd')
    assert Command('cmd') != Command('other')

    assert Command([1, 2, 3]) == Command([1, 2, 3])
    assert Command([1, 2, 3]) != Command([11, 22, 33])

    assert Command('arb') != 'arb'

    assert Command('cmd',
                   is_shell=True,
                   cwd='cwd',
                   is_save=False,
                   is_text=True,
                   stdout='out',
                   stderr='err',
                   encoding='enc',
                   append=True) == Command('cmd',
                                           is_shell=True,
                                           cwd='cwd',
                                           is_save=False,
                                           is_text=True,
                                           stdout='out',
                                           stderr='err',
                                           encoding='enc',
                                           append=True)

    assert Command('cmd',
                   is_shell=False,
                   cwd='cwd',
                   is_save=False,
                   is_text=True,
                   stdout='out',
                   stderr='err',
                   encoding='enc',
                   append=False) != Command('cmd',
                                            is_shell=False,
                                            cwd='cwd',
                                            is_save=False,
                                            is_text=True,
                                            stdout='out',
                                            stderr='err',
                                            encoding='enc',
                                            append=True)
    assert Command('cmd',
                   is_shell=True,
                   cwd='cwd',
                   is_save=False,
                   is_text=True,
                   stdout='out',
                   stderr='err',
                   encoding='enc1',
                   append=True) != Command('cmd',
                                           is_shell=True,
                                           cwd='cwd',
                                           is_save=False,
                                           is_text=True,
                                           stdout='out',
                                           stderr='err',
                                           encoding='enc2',
                                           append=True)
    assert Command('cmd',
                   is_shell=True,
                   cwd='cwd',
                   is_save=True,
                   is_text=True,
                   encoding='enc',
                   append=True) == Command('cmd',
                                           is_shell=True,
                                           cwd='cwd',
                                           is_save=True,
                                           is_text=True,
                                           encoding='enc',
                                           append=True)

    assert [Command('one')] == [Command('one')]
    assert [Command('one')] != [Command('two')]

    # results
    cmd1 = Command('cmd')
    cmd1.results.append('one')
    cmd2 = Command('cmd')
    cmd2.results.append('one')
    assert cmd1 == cmd2
    cmd2.results.append('two')
    assert cmd1 != cmd2


def test_subproc_command_no_stdout_on_save():
    """Raise when combining stdout/stdour with save."""
    #  stdout
    with pytest.raises(ContextError) as err:
        Command('arb', is_save=True, stdout='in')

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")

    #  stderr
    with pytest.raises(ContextError) as err:
        Command('arb', is_save=True, stderr='out')

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")

    #  stdout + stdout
    with pytest.raises(ContextError) as err:
        Command('arb', is_save=True, stdout='in', stderr='out')

    assert str(err.value) == (
        "You can't set `stdout` or `stderr` when `save` is True.")


def test_subproc_command_output_handles_none():
    """Output handles return None."""
    cmd = Command('arb')
    with cmd.output_handles() as (stdout, stderr):
        assert stdout is None
        assert stderr is None


def test_subproc_command_output_handles_dev_null():
    """Output handles to /dev/null."""
    cmd = Command('arb', stdout='/dev/null', stderr='/dev/null')
    with cmd.output_handles() as (stdout, stderr):
        assert stdout == subprocess.DEVNULL
        assert stderr == subprocess.DEVNULL

    cmd = Command('arb', stderr='/dev/null')
    with cmd.output_handles() as (stdout, stderr):
        assert stdout is None
        assert stderr == subprocess.DEVNULL

    cmd = Command('arb', stdout='/dev/null')
    with cmd.output_handles() as (stdout, stderr):
        assert stdout == subprocess.DEVNULL
        assert stderr is None

    cmd = Command('arb', stdout='/dev/null', stderr='/dev/stdout')
    with cmd.output_handles() as (stdout, stderr):
        assert stdout == subprocess.DEVNULL
        assert stderr == subprocess.STDOUT
# endregion Command

# region SubprocessResult


def test_subprocessresult_check_returncode_zero():
    """Return None when returncode is 0."""
    sr = SubprocessResult('mycmd', 0, stdout='out', stderr='err')
    assert sr.check_returncode() is None


def test_subprocessresult_check_returncode_nonzero():
    """Return SubprocessError when returncode is non-zero."""
    sr = SubprocessResult('mycmd', 1, stdout='out', stderr='err')
    err = sr.check_returncode()
    assert type(err) is SubprocessError
    assert err.cmd == 'mycmd'
    assert err.returncode == 1
    assert err.stdout == 'out'
    assert err.stderr == 'err'

    sr = SubprocessResult('mycmd', -1, stdout='out', stderr='err')
    err = sr.check_returncode()
    assert type(err) is SubprocessError
    assert err.cmd == 'mycmd'
    assert err.returncode == -1
    assert err.stdout == 'out'
    assert err.stderr == 'err'

    sr = SubprocessResult('mycmd', 2)
    err = sr.check_returncode()
    assert type(err) is SubprocessError
    assert err.cmd == 'mycmd'
    assert err.returncode == 2
    assert err.stdout is None
    assert err.stderr is None


def test_subprocessresult_dict():
    """Get items like dict from SubprocessResult. For backwards compat."""
    sr = SubprocessResult('mycmd', 123, stdout='out', stderr='err')
    assert sr['cmd'] == 'mycmd'
    assert sr['returncode'] == 123
    assert sr['stdout'] == 'out'
    assert sr['stderr'] == 'err'

    # works as a formatting expression
    context = Context({'cmdOut': sr})
    out_str = context.get_formatted_value(
        'begin {cmdOut.returncode} {cmdOut.stdout} {cmdOut.stderr} end')
    assert out_str == 'begin 123 out err end'

    out_str = context.get_formatted_value(
        'begin {cmdOut[returncode]} {cmdOut[stdout]} {cmdOut[stderr]} end')
    assert out_str == 'begin 123 out err end'

    # works in a PyString
    assert context.get_formatted_value(PyString("cmdOut.returncode")) == 123
    assert context.get_formatted_value(PyString("cmdOut['returncode']")) == 123


def test_subprocessresult_repr():
    """Convert between SubprocessResult repr and instance."""
    sr = repr(SubprocessResult('mycmd', 123))
    rehydrated = eval(sr)
    assert rehydrated.cmd == 'mycmd'
    assert rehydrated.returncode == 123
    assert rehydrated.stdout is None
    assert rehydrated.stderr is None

    sr = repr(SubprocessResult('mycmd', 123, stdout=b'stdout'))
    rehydrated = eval(sr)
    assert rehydrated.cmd == 'mycmd'
    assert rehydrated.returncode == 123
    assert rehydrated.stdout == b'stdout'
    assert rehydrated.stderr is None

    sr = repr(SubprocessResult('mycmd', 123, stderr='err'))
    rehydrated = eval(sr)
    assert rehydrated.cmd == 'mycmd'
    assert rehydrated.returncode == 123
    assert rehydrated.stdout is None
    assert rehydrated.stderr == 'err'

    sr = repr(SubprocessResult('mycmd', 123, stdout='out', stderr='err'))
    rehydrated = eval(sr)
    assert rehydrated.cmd == 'mycmd'
    assert rehydrated.returncode == 123
    assert rehydrated.stdout == 'out'
    assert rehydrated.stderr == 'err'

    sr = repr(SubprocessResult(['my', 'cmd'], -123,
                               stdout=b'out', stderr=b'err'))
    rehydrated = eval(sr)
    assert rehydrated.cmd == ['my', 'cmd']
    assert rehydrated.returncode == -123
    assert rehydrated.stdout == b'out'
    assert rehydrated.stderr == b'err'


def test_subprocessresult_str():
    """Convert SubprocessResult to friendly string."""
    sr = str(SubprocessResult('mycmd', 123))
    assert sr == """\
cmd: mycmd
returncode: 123
stdout: None
stderr: None
"""

    sr = str(SubprocessResult('mycmd', 123, stdout=b'out'))
    assert sr == """\
cmd: mycmd
returncode: 123
stdout: b'out'
stderr: None
"""

    sr = str(SubprocessResult('mycmd', 123, stderr=b'err'))
    assert sr == """\
cmd: mycmd
returncode: 123
stdout: None
stderr: b'err'
"""

    sr = str(SubprocessResult('mycmd', 123, stdout=b'out', stderr='err'))
    assert sr == """\
cmd: mycmd
returncode: 123
stdout: b'out'
stderr: err
"""

    sr = str(SubprocessResult(['my', 'cmd'], -123,
                              stdout='out', stderr='err'))
    assert sr == """\
cmd: ['my', 'cmd']
returncode: -123
stdout: out
stderr: err
"""
# endregion SubprocessResult



"""pypyr venv step yaml definition classes - domain specific language."""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
from collections.abc import Mapping, Sequence
from concurrent.futures import ThreadPoolExecutor
import logging

from pypyr.context import Context
from pypyr.errors import ContextError, MultiError
from pypyr.venv import VenvCreator, SIMPLE_INPUT_TYPE

logger = logging.getLogger(__name__)

EXCEPTION_INPUT = """venv input is wrong. Your input is {bad_type}.
It should be either:

venv: ./path/mydir

OR to specify non-default options (expanded syntax):
venv:
  path: ./path/mydir
  pip: package1 package2
  clear: True

OR to create multiple venvs with default options:
venv:
  - ./path1/dir1
  - /path2/dir3

(Each list item can be a simple string or a mapping in expanded syntax.)

OR to create multiple venvs with the same non-default options:
venv:
  path:
    - ./path1/dir1
    - /path2/dir2
  pip: package1 package2
  clear: True"""

EXCEPTION_LIST_INPUT = """venv list input is wrong. Your input is {bad_type}.
Each list item should be a simple string or a mapping.

venv:
  - ./path1/dir1
  - /path2/dir3

OR to specify non-default options in expanded syntax:
venv:
  - path: ./path1/dir1
    pip: package1 package2
    clear: True

  - path: /path2/dir2
    pip: package3 package4

OR you can mix simple strings for default options and mappings for non-default:
venv:
  - ./path1/dir1

  - path: ./path2/dir2
    pip: package1 package2
    clear: True
"""


class VenvCreatorStep():
    """Model the venv create built-in step and do the actual creation work."""

    CONFIG_KEY = 'venv'

    def __init__(self, venvs: list[VenvCreator]):
        """Initialize step."""
        self.venvs: list[VenvCreator] = venvs

    @classmethod
    def from_context(cls, context: Context) -> VenvCreatorStep:
        """Initialize step from context using this factory."""
        context.assert_key_has_value(key=VenvCreatorStep.CONFIG_KEY,
                                     caller='pypyr.steps.venv')

        venv_config = context.get_formatted(VenvCreatorStep.CONFIG_KEY)

        venvs = []
        if isinstance(venv_config, SIMPLE_INPUT_TYPE):
            venvs.append(VenvCreator(venv_config))
        elif isinstance(venv_config, Mapping):
            venvs.extend(VenvCreator.from_mapping(venv_config))
        elif isinstance(venv_config, Sequence):
            for venv in venv_config:
                if isinstance(venv, SIMPLE_INPUT_TYPE):
                    venvs.append(VenvCreator(venv))
                elif isinstance(venv, Mapping):
                    venvs.extend(VenvCreator.from_mapping(venv))
                else:
                    raise ContextError(
                        EXCEPTION_LIST_INPUT.format(
                            bad_type=type(venv).__name__))
        else:
            raise ContextError(
                EXCEPTION_INPUT.format(bad_type=type(venv_config).__name__))

        return cls(venvs)

    def run_step(self) -> None:
        """Create the venvs when the step runs."""
        logger.debug("started")

        with ThreadPoolExecutor() as executor:
            for venv in self.venvs:
                venv.create_in_executor(executor)

        logger.debug("venv creation threadpool joined...checking results")
        # after all the venv creation is done, can install pip deps
        # pip shouldn't run in parallel, there be dragons.
        errors: list[Exception] = []
        for venv in self.venvs:
            result = venv.check_result()
            # if result contains anything it's an exception
            if result:
                logger.error("venv creation failed for %s with %s",
                             venv.path, result)
                errors.append(result)
            else:
                # install pip deps only if env created successfully
                try:
                    if venv.with_pip:
                        venv.install_dependencies()
                    else:
                        logger.debug(
                            "Not installing any extra dependencies because "
                            + "with_pip is False for %s.", venv.path)

                    logger.info("created venv %s", venv.path)
                except Exception as ex:
                    logger.error("failed installing dependencies to venv %s",
                                 venv.path)
                    errors.append(ex)

        if errors:
            if len(errors) == 1:
                raise errors[0]
            else:
                raise MultiError("venvs creation failed:", errors)

        logger.debug("done")



"""call.py unit tests."""
import logging
import pytest
from pypyr.context import Context
from pypyr.errors import Call
from pypyr.steps.call import run_step

from tests.common.utils import patch_logger


def test_call_step_dict_with_all_args():
    """Dict with all values set."""
    with pytest.raises(Call) as err:
        with patch_logger('pypyr.steps.call',
                          logging.INFO) as mock_logger_info:
            run_step(Context(Context({'call': {'groups': ['b', 'c'],
                                               'success': 'sg',
                                               'failure': 'fg'}})))

    cof = err.value
    assert isinstance(cof, Call)
    assert cof.groups == ['b', 'c']
    assert cof.success_group == 'sg'
    assert cof.failure_group == 'fg'
    assert cof.original_config == ('call', {'groups': ['b', 'c'],
                                            'success': 'sg',
                                            'failure': 'fg'})

    mock_logger_info.assert_called_once_with(
        "step pypyr.steps.call about to hand over control with call: "
        "Will run groups: ['b', 'c']  with success sg and failure fg")



"""Toml handling."""
try:
    # can get rid of the try soon as py 3.11 min supported version
    # reason is py 3.11 includes tomli in stdlib
    import tomllib as toml_reader
except ModuleNotFoundError:  # pragma: no cover
    import tomli as toml_reader

import tomli_w as toml_writer  # type: ignore


def read_file(path):
    """Read toml file at path and return object.

    Args:
        path (Path-like): Read file from this location.

    Returns:
        dict.
    """
    with open(path, 'rb') as file:
        return load(file)


def write_file(path, payload):
    """Write payload to path as a toml file.

    payload should not be None.

    Args:
        path (Path-like): Write file to this location.
        payload (dict-like): Serialize this object to toml string & write to
                             path.

    Returns:
        None.
    """
    with open(path, 'wb') as file:
        dump(payload, file)


def load(file):
    """Read open file object as toml.

    File-like MUST be open in binary mode.

    Args:
        file (file-like): Open file-like object.

    Returns:
        dict
    """
    return toml_reader.load(file)


def dump(payload, file):
    """Write payload to open file-like object.

    File-like MUST be open in binary mode.

    Args:
        file (file-like): Open file-like object.
        payload (dict-like): Serialize this object to toml string & write to
                             file.

    Returns:
        None.
    """
    toml_writer.dump(payload, file)



"""poll.py unit tests."""
import logging
from unittest.mock import call, MagicMock, patch
import pypyr.utils.poll as poll


# region wait_until_true
from tests.common.utils import patch_logger


@patch('time.sleep')
def test_wait_until_true_with_static_decorator(mock_time_sleep):
    """wait_until_true with static decorator."""
    mock = MagicMock()
    mock.side_effect = [
        'test string 1',
        'test string 2',
        'test string 3',
        'expected value',
        'test string 5'
    ]

    @poll.wait_until_true(interval=0.01, max_attempts=10)
    def decorate_me(arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        return mock(arg1) == 'expected value'

    assert decorate_me('v1', 'v2')
    assert mock.call_count == 4
    mock.assert_called_with('v1')
    assert mock_time_sleep.call_count == 3
    mock_time_sleep.assert_called_with(0.01)


@patch('time.sleep')
def test_wait_until_true_invoke_inline(mock_time_sleep):
    """wait_until_true with dynamic invocation."""
    mock = MagicMock()
    mock.side_effect = [
        'test string 1',
        'test string 2',
        'test string 3',
        'expected value',
        'test string 5'
    ]

    def decorate_me(arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        return mock(arg1) == 'expected value'

    assert poll.wait_until_true(interval=0.01, max_attempts=10)(
        decorate_me)('v1', 'v2')
    assert mock.call_count == 4
    mock.assert_called_with('v1')
    assert mock_time_sleep.call_count == 3
    mock_time_sleep.assert_called_with(0.01)


@patch('time.sleep')
def test_wait_until_true_with_timeout(mock_time_sleep):
    """wait_until_true with dynamic invocation, exhaust wait attempts."""
    mock = MagicMock()
    mock.side_effect = [
        'test string 1',
        'test string 2',
        'test string 3',
        'test string 4',
        'test string 5',
        'test string 6',
        'test string 7',
        'test string 8',
        'test string 9',
        'test string 10',
        'test string 11',
    ]

    def decorate_me(arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        return mock(arg1) == 'expected value'

    assert not poll.wait_until_true(interval=0.01, max_attempts=10)(
        decorate_me)('v1', 'v2')
    assert mock.call_count == 10
    mock.assert_called_with('v1')
    assert mock_time_sleep.call_count == 9
    mock_time_sleep.assert_called_with(0.01)


@patch('time.sleep')
def test_wait_until_true_once_not_found(mock_time_sleep):
    """wait_until_true max_attempts 1."""
    mock = MagicMock()
    mock.side_effect = [
        'test string 1',
        'test string 2',
    ]

    def decorate_me(arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        return mock(arg1) == 'expected value'

    assert not poll.wait_until_true(interval=0.01, max_attempts=1)(
        decorate_me)('v1', 'v2')
    mock.assert_called_once_with('v1')
    mock_time_sleep.assert_not_called()


@patch('time.sleep')
def test_wait_until_true_once_found(mock_time_sleep):
    """wait_until_true max_attempts 1."""
    mock = MagicMock()
    mock.side_effect = [
        'expected value',
        'test string 2',
    ]

    def decorate_me(arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        return mock(arg1) == 'expected value'

    assert poll.wait_until_true(interval=0.01, max_attempts=1)(
        decorate_me)('v1', 'v2')
    mock.assert_called_once_with('v1')
    mock_time_sleep.assert_not_called()

# endregion wait_until_true

# region while_until_true


@patch('time.sleep')
def test_while_until_true_with_static_decorator(mock_time_sleep):
    """while_until_true with static decorator."""
    mock = MagicMock()
    mock.side_effect = [
        'test string 1',
        'test string 2',
        'test string 3',
        'expected value',
        'test string 5'
    ]

    actual_counter = 0

    @poll.while_until_true(interval=0.01, max_attempts=10)
    def decorate_me(counter, arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        nonlocal actual_counter
        actual_counter += 1
        assert actual_counter == counter
        return mock(arg1) == 'expected value'

    # the decorator injects the 1st arg (counter)
    assert decorate_me('v1', 'v2')
    assert mock.call_count == 4
    mock.assert_called_with('v1')
    assert mock_time_sleep.mock_calls == [call(0.01), call(0.01), call(0.01)]


@patch('time.sleep')
def test_while_until_true_invoke_inline(mock_time_sleep):
    """while_until_true with dynamic invocation."""
    mock = MagicMock()
    mock.side_effect = [
        'test string 1',
        'test string 2',
        'test string 3',
        'expected value',
        'test string 5'
    ]

    actual_counter = 0

    def decorate_me(counter, arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        nonlocal actual_counter
        actual_counter += 1
        assert actual_counter == counter
        return mock(arg1) == 'expected value'

    assert poll.while_until_true(interval=0.01, max_attempts=10)(
        decorate_me)('v1', 'v2')
    assert mock.call_count == 4
    mock.assert_called_with('v1')
    assert mock_time_sleep.mock_calls == [call(0.01), call(0.01), call(0.01)]


@patch('time.sleep')
def test_while_until_true_with_exhaust(mock_time_sleep):
    """while_until_true with dynamic invocation, exhaust wait attempts."""
    mock = MagicMock()
    mock.side_effect = [
        'test string 1',
        'test string 2',
        'test string 3',
        'test string 4',
        'test string 5',
        'test string 6',
        'test string 7',
        'test string 8',
        'test string 9',
        'test string 10',
        'test string 11',
    ]

    actual_counter = 0

    def decorate_me(counter, arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        nonlocal actual_counter
        actual_counter += 1
        assert actual_counter == counter
        out = mock(arg1)
        assert out == f'test string {counter}'
        return out == 'expected value'

    assert not poll.while_until_true(interval=0.01, max_attempts=10)(
        decorate_me)('v1', 'v2')
    assert mock.call_count == 10
    mock.assert_called_with('v1')
    assert mock_time_sleep.call_count == 9
    mock_time_sleep.assert_called_with(0.01)


@patch('time.sleep')
def test_while_until_true_once_not_found(mock_time_sleep):
    """while_until_true max_attempts 1."""
    mock = MagicMock()
    mock.side_effect = [
        'test string 1',
        'test string 2',
    ]

    actual_counter = 0

    def decorate_me(counter, arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        nonlocal actual_counter
        actual_counter += 1
        assert actual_counter == counter
        return mock(arg1) == 'expected value'

    assert not poll.while_until_true(interval=0.01, max_attempts=1)(
        decorate_me)('v1', 'v2')
    mock.assert_called_once_with('v1')
    mock_time_sleep.assert_not_called()


@patch('time.sleep')
def test_while_until_true_once_found(mock_time_sleep):
    """wait_until_true max_attempts 1."""
    mock = MagicMock()
    mock.side_effect = [
        'expected value',
        'test string 2',
    ]

    actual_counter = 0

    def decorate_me(counter, arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        nonlocal actual_counter
        actual_counter += 1
        assert actual_counter == counter
        return mock(arg1) == 'expected value'

    assert poll.while_until_true(interval=0.01, max_attempts=1)(
        decorate_me)('v1', 'v2')
    mock.assert_called_once_with('v1')
    mock_time_sleep.assert_not_called()


@patch('time.sleep')
def test_while_until_true_no_max(mock_time_sleep):
    """while_until_true with dynamic invocation, infinite (max is None)."""
    mock = MagicMock()
    mock.side_effect = [
        'test string 1',
        'test string 2',
        'test string 3',
        'test string 4',
        'test string 5',
        'test string 6',
        'test string 7',
        'test string 8',
        'test string 9',
        'test string 10',
        'test string 11',
    ]

    actual_counter = 0

    def decorate_me(counter, arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        nonlocal actual_counter
        actual_counter += 1
        assert actual_counter == counter
        out = mock(arg1)
        assert out == f'test string {counter}'
        return out == 'test string 11'

    with patch_logger('pypyr.utils.poll', logging.DEBUG) as mock_logger_debug:
        assert (poll.while_until_true(interval=0.01,
                                      max_attempts=None)(decorate_me)('v1',
                                                                      'v2'))
    assert mock_logger_debug.mock_calls == [
        call('started'),
        call('Looping every 0.01 seconds.'),
        call('iteration 1. Sleeping for 0.01s. Still waiting...'),
        call('iteration 2. Sleeping for 0.01s. Still waiting...'),
        call('iteration 3. Sleeping for 0.01s. Still waiting...'),
        call('iteration 4. Sleeping for 0.01s. Still waiting...'),
        call('iteration 5. Sleeping for 0.01s. Still waiting...'),
        call('iteration 6. Sleeping for 0.01s. Still waiting...'),
        call('iteration 7. Sleeping for 0.01s. Still waiting...'),
        call('iteration 8. Sleeping for 0.01s. Still waiting...'),
        call('iteration 9. Sleeping for 0.01s. Still waiting...'),
        call('iteration 10. Sleeping for 0.01s. Still waiting...'),
        call('iteration 11. Desired state reached.'),
        call('done')]

    assert mock.call_count == 11
    mock.assert_called_with('v1')
    assert mock_time_sleep.call_count == 10
    mock_time_sleep.assert_called_with(0.01)


@patch('time.sleep')
def test_while_until_true_max_exhaust(mock_time_sleep):
    """while_until_true with dynamic invocation, exhaust max."""
    mock = MagicMock()
    mock.side_effect = [
        'test string 1',
        'test string 2',
        'test string 3',
    ]

    actual_counter = 0

    def decorate_me(counter, arg1, arg2):
        """Test static decorator syntax."""
        assert arg1 == 'v1'
        assert arg2 == 'v2'
        nonlocal actual_counter
        actual_counter += 1
        assert actual_counter == counter
        out = mock(arg1)
        assert out == f'test string {counter}'
        return False

    with patch_logger('pypyr.utils.poll', logging.DEBUG) as mock_logger_debug:
        assert not (poll.while_until_true(interval=0.01,
                                          max_attempts=3)(decorate_me)('v1',
                                                                       'v2'))
    assert mock_logger_debug.mock_calls == [
        call('started'),
        call('Looping every 0.01 seconds for 3 attempts'),
        call('iteration 1. Sleeping for 0.01s. Still waiting...'),
        call('iteration 2. Sleeping for 0.01s. Still waiting...'),
        call('iteration 3. Max attempts exhausted.'),
        call('done')]

    assert mock.call_count == 3
    mock.assert_called_with('v1')
    assert mock_time_sleep.call_count == 2
    assert mock_time_sleep.mock_calls == [call(0.01), call(0.01)]


def arb_callable(n):
    """Arb callable for testing."""
    return n * 2


@patch('time.sleep')
def test_while_until_true_callable_with_max_attempts(mock_sleep):
    """Interval calculated from callable per iteration with max set."""
    decorate_me = MagicMock()
    decorate_me.side_effect = [False, False, True, False]

    assert poll.while_until_true(interval=arb_callable,
                                 max_attempts=3)(decorate_me)('arg1', 'arg2')

    assert decorate_me.mock_calls == [call(1, 'arg1', 'arg2'),
                                      call(2, 'arg1', 'arg2'),
                                      call(3, 'arg1', 'arg2')]

    assert mock_sleep.mock_calls == [call(2), call(4)]


@patch('time.sleep')
def test_while_until_true_callable_without_max_attempts(mock_sleep):
    """Interval calculated from callable per iteration infinite."""
    decorate_me = MagicMock()
    decorate_me.side_effect = [False, False, False, True, False]

    assert poll.while_until_true(interval=arb_callable,
                                 max_attempts=None)(decorate_me)('arg1',
                                                                 'arg2')

    assert decorate_me.mock_calls == [call(1, 'arg1', 'arg2'),
                                      call(2, 'arg1', 'arg2'),
                                      call(3, 'arg1', 'arg2'),
                                      call(4, 'arg1', 'arg2')]

    assert mock_sleep.mock_calls == [call(2), call(4), call(6)]
# endregion while_until_true



"""pypyr.steps.envget unit tests."""
import logging
import os
import pytest
from pypyr.context import Context
from pypyr.errors import ContextError, KeyNotInContextError
import pypyr.steps.envget

# ------------------------- run_step -----------------------------------------#
from tests.common.utils import patch_logger


def test_envget_throws_on_empty_context():
    """Context must exist."""
    with pytest.raises(AssertionError) as err_info:
        pypyr.steps.envget.run_step(None)

    assert str(err_info.value) == ("context must have value "
                                   "for pypyr.steps.envget")


def test_envget_throws_if_envget_context_missing():
    """Env context keys must exist."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.envget.run_step(Context({'arbkey': 'arbvalue'}))

    assert str(err_info.value) == ("context['envGet'] doesn't exist. It must "
                                   "exist for pypyr.steps.envget.")


def test_envget_throws_if_env_context_wrong_type():
    """Env context keys must exist."""
    with pytest.raises(ContextError) as err_info:
        pypyr.steps.envget.run_step(
            Context({'envGet': 'it shouldnt be a string'}))

    assert str(err_info.value) == ("envGet must contain a list of dicts.")


def test_envget_pass():
    """Env get success case."""
    os.environ['ARB_DELETE_ME1'] = 'arb value from $ENV ARB_DELETE_ME1'
    os.environ['ARB_DELETE_ME2'] = 'arb value from $ENV ARB_DELETE_ME2'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'envGet': [
            {'env': 'ARB_DELETE_ME1',
             'key': 'key2',
             'default': 'blah'},
            {'env': 'ARB_DELETE_ME2',
             'key': 'key3'},
            {'env': 'ARB_XXX_IDONTEXIST_BLAH',
             'key': 'key4',
             'default': 'blah4'},
            {'env': 'ARB_XXX_IDONTEXIST_BLAH',
             'key': 'key5'}
        ]
    })

    with patch_logger('pypyr.steps.envget', logging.INFO) as mock_logger_info:
        pypyr.steps.envget.run_step(context)

    del os.environ['ARB_DELETE_ME1']
    del os.environ['ARB_DELETE_ME2']

    mock_logger_info.assert_called_once_with('saved 3 $ENVs to context.')

    assert context['key1'] == 'value1'
    assert context['key2'] == 'arb value from $ENV ARB_DELETE_ME1'
    assert context['key3'] == 'arb value from $ENV ARB_DELETE_ME2'
    assert context['key4'] == 'blah4'
    assert 'key5' not in context


def test_envget_pass_single_item():
    """Env get success case where single item not list passed."""
    os.environ['ARB_DELETE_ME1'] = 'arb value from $ENV ARB_DELETE_ME1'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'envGet':
            {'env': 'ARB_DELETE_ME1',
             'key': 'key2',
             'default': 'blah'}
    })

    pypyr.steps.envget.run_step(context)

    del os.environ['ARB_DELETE_ME1']

    assert context['key1'] == 'value1'
    assert context['key2'] == 'arb value from $ENV ARB_DELETE_ME1'
    assert context['key3'] == 'value3'
    assert 'key5' not in context


def test_envget_pass_nones():
    """Env get success case with none/nulls."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'envGet': [
            {'env': 'ARB_XXX_IDONTEXIST_BLAH',
             'key': 'key4',
             'default': None}
        ]
    })

    pypyr.steps.envget.run_step(context)

    assert context['key4'] is None


def test_envget_pass_with_substitutions():
    """Env get success case with string substitutions."""
    os.environ['ARB_DELETE_ME1'] = 'arb value from $ENV ARB_DELETE_ME1'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'env_val1': 'ARB_DELETE_ME1',
        'env_val2': 'ARB_DELETE_ME2',
        'default_val': 'blah',
        'key_val': 'key3',
        'envGet': [
            {'env': '{env_val1}',
             'key': '{key_val}',
             'default': 'blah'},
            {'env': '{env_val2}',
             'key': 'key4',
             'default': '{default_val}'}
        ]
    })

    pypyr.steps.envget.run_step(context)

    del os.environ['ARB_DELETE_ME1']

    assert context['key1'] == 'value1'
    assert context['key2'] == 'value2'
    assert context['key3'] == 'arb value from $ENV ARB_DELETE_ME1'
    assert context['key4'] == 'blah'


def test_envget_pass_with_substitutions_default_not_string():
    """Env get success case with substitutions that aren't string."""
    os.environ['ARB_DELETE_ME1'] = 'arb value from $ENV ARB_DELETE_ME1'

    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'env_val1': 'ARB_DELETE_ME1',
        'env_val2': 'ARB_DELETE_ME2',
        'default_val': [0, 1, 2],
        'key_val': 'key3',
        'key_key': 'key',
        'envGet': [
            {'env': '{env_val1}',
             '{key_key}': '{key_val}',
             'default': 'blah'},
            {'env': '{env_val2}',
             'key': 'key4',
             'default': '{default_val}'}
        ]
    })

    pypyr.steps.envget.run_step(context)

    del os.environ['ARB_DELETE_ME1']

    assert context['key1'] == 'value1'
    assert context['key2'] == 'value2'
    assert context['key3'] == 'arb value from $ENV ARB_DELETE_ME1'
    assert context['key4'] == [0, 1, 2]

# ------------------------- get_args -----------------------------------------#


def test_get_args_not_a_dict():
    """Env context keys must exist."""
    with pytest.raises(ContextError) as err_info:
        pypyr.steps.envget.get_args('it shouldnt be a string')

    assert str(err_info.value) == ("envGet must contain a list of dicts.")


def test_get_args_env_exists():
    """Env must exist."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.envget.get_args({'blah': 'blah'})

    assert str(err_info.value) == (
        "context envGet[env] must exist in context for envGet.")


def test_get_args_empty_env_raises():
    """Env must not be empty."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.envget.get_args({'env': None})

    assert str(err_info.value) == (
        "context envGet[env] must exist in context for envGet.")


def test_get_args_key_exists():
    """Key must exist."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.envget.get_args({'env': 'v'})

    assert str(err_info.value) == (
        "context envGet[key] must exist in context for envGet.")


def test_get_args_empty_key_raises():
    """Empty key must raise."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.envget.get_args({'env': 'v', 'key': ''})

    assert str(err_info.value) == (
        "context envGet[key] must exist in context for envGet.")


def test_get_args_no_default():
    """Parses with no default set."""
    (env, key, has_default, default) = pypyr.steps.envget.get_args(
        {'env': 'e', 'key': 'k'})

    assert env == 'e'
    assert key == 'k'
    assert not has_default
    assert default is None


def test_get_args_with_default():
    """Parses with default set."""
    (env, key, has_default, default) = pypyr.steps.envget.get_args(
        {'env': 'e', 'key': 'k', 'default': 'blah'})

    assert env == 'e'
    assert key == 'k'
    assert has_default
    assert default == 'blah'


def test_get_args_with_none_default():
    """Parses with none default set."""
    (env, key, has_default, default) = pypyr.steps.envget.get_args(
        {'env': 'e', 'key': 'k', 'default': None})

    assert env == 'e'
    assert key == 'k'
    assert has_default
    assert default is None
# ------------------------- END get_args -------------------------------------#
